#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2014 Hamilton Kibbe <ham@hamiltonkib.be>
# Copyright 2022 Jan Sebastian Götte <gerbonara@jaseg.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import io
import sys
import re
import warnings
import copy
import bisect
import textwrap
import itertools
from collections import namedtuple
from pathlib import Path
from zipfile import ZipFile, is_zipfile
from collections import defaultdict
import tempfile

from .excellon import ExcellonFile, parse_allegro_ncparam, parse_allegro_logfile
from .rs274x import GerberFile
from .ipc356 import Netlist
from .cam import FileSettings, LazyCamFile
from .layer_rules import MATCH_RULES
from .utils import sum_bounds, setup_svg, MM, Tag, convex_hull
from . import graphic_objects as go
from . import graphic_primitives as gp


STANDARD_LAYERS = [
        'mechanical outline',
        'top copper',
        'top mask',
        'top silk',
        'top paste',
        'bottom copper',
        'bottom mask',
        'bottom silk',
        'bottom paste',
        ]

DEFAULT_COLORS = {
        'copper': '#cccccc',
        'mask': '#004200bf',
        'paste': '#999999',
        'silk': '#e0e0e0',
        'drill': '#303030',
        'outline': '#F0C000',
    }

class NamingScheme:
    kicad = {
    'top copper':           '{board_name}-F.Cu.gbr',
    'top mask':             '{board_name}-F.Mask.gbr',
    'top silk':             '{board_name}-F.SilkS.gbr',
    'top paste':            '{board_name}-F.Paste.gbr',
    'bottom copper':        '{board_name}-B.Cu.gbr',
    'bottom mask':          '{board_name}-B.Mask.gbr',
    'bottom silk':          '{board_name}-B.SilkS.gbr',
    'bottom paste':         '{board_name}-B.Paste.gbr',
    'inner copper':         '{board_name}-In{layer_number}.Cu.gbr',
    'mechanical outline':   '{board_name}-Edge.Cuts.gbr',
    'drill unknown':        '{board_name}.drl',
    'drill plated':         '{board_name}-PTH.drl',
    'drill nonplated':      '{board_name}-NPTH.drl',
    'other comments':       '{board_name}-Cmts.User.gbr',
    'other drawings':       '{board_name}-Dwgs.User.gbr',
    'top fabrication':      '{board_name}-F.Fab.gbr',
    'bottom fabrication':   '{board_name}-B.Fab.gbr',
    'top adhesive':         '{board_name}-F.Adhes.gbr',
    'bottom adhesive':      '{board_name}-B.Adhes.gbr',
    'top courtyard':        '{board_name}-F.CrtYd.gbr',
    'bottom courtyard':     '{board_name}-B.CrtYd.gbr',
    'other netlist':        '{board_name}.d356',
    }

    altium = {
    'top copper':           '{board_name}.gtl',
    'top mask':             '{board_name}.gts',
    'top silk':             '{board_name}.gto',
    'top paste':            '{board_name}.gtp',
    'bottom copper':        '{board_name}.gbl',
    'bottom mask':          '{board_name}.gbs',
    'bottom silk':          '{board_name}.gbo',
    'bottom paste':         '{board_name}.gbp',
    'inner copper':         '{board_name}.gp{layer_number}',
    'mechanical outline':   '{board_name}.gko',
    'drill unknown':        '{board_name}.drl',
    'drill plated':         '{board_name}.plated.drl',
    'drill nonplated':      '{board_name}.nonplated.drl',
    'other comments':       '{board_name}.gm2',
    'other drawings':       '{board_name}.gm3',
    'top courtyard':        '{board_name}.gm13',
    'bottom courtyard':     '{board_name}.gm14',
    'top fabrication':      '{board_name}.gm15',
    'bottom fabrication':   '{board_name}.gm16',
    }



def _match_files(filenames):
    matches = {}
    for generator, rules in MATCH_RULES.items():
        already_matched = set()
        gen = {}
        matches[generator] = gen
        for layer, regex in rules.items():
            for fn in filenames:
                if fn in already_matched:
                    continue

                if (m := re.fullmatch(regex, fn.name, re.IGNORECASE)):
                    if layer == 'inner copper':
                        target = 'inner_' + ''.join(e or '' for e in m.groups()) + ' copper'
                    else:
                        target = layer

                    gen[target] = gen.get(target, []) + [fn]
                    already_matched.add(fn)
    return matches


def _best_match(filenames):
    matches = _match_files(filenames)
    matches = sorted(matches.items(), key=lambda pair: len(pair[1]))
    generator, files = matches[-1]
    return generator, files


def identify_file(data):
    """ Identify file type from file contents. Returns either of the string constants :py:obj:`excellon`,
    :py:obj:`gerber`, or :py:obj:`ipc356`, or returns :py:obj:`None` if the file format is unclear.

    :param data: Contents of file as :py:obj:`str`
    :rtype: :py:obj:`str`
    """

    if 'M48' in data:
        return 'excellon'

    if 'G90' in data and ';LEADER:' in data: # yet another allegro special case
        return 'excellon'

    if 'FSLAX' in data or 'FSTAX' in data:
        return 'gerber'

    if 'UNITS CUST' in data:
        return 'ipc356'

    return None


def _common_prefix(l):
    out = []
    for cand in l:
        score = lambda n: sum(elem.startswith(cand[:n]) for elem in l)
        baseline = score(1)
        if len(l) - baseline > 5:
            continue
        for n in range(len(cand) if '.' not in cand else cand.index('.')+1, 2, -1):
            if len(l) - score(n) < 5:
                break
        out.append(cand[:n-1])
 
    if not out:
        return ''
 
    return sorted(out, key=len)[-1]

def _do_autoguess(filenames):
    prefix = _common_prefix([f.name for f in filenames])

    matches = {}
    for f in filenames:
        name = _layername_autoguesser(f.name[len(prefix):] if f.name.startswith(prefix) else f.name)
        if name != 'unknown unknown':
            matches[name] = matches.get(name, []) + [f]

    inner_layers = [ m for m in matches if 'inner' in m ]
    if len(inner_layers) >= 2 and 'copper top' not in matches and 'copper bottom' not in matches:
        if 'inner_01 copper' in matches:
            warnings.warn('Could not find copper layer. Re-assigning outermost inner layers to top/bottom copper.')
            matches['top copper'] = matches.pop('inner_01 copper')
            last_inner = sorted(inner_layers, key=lambda name: int(name.partition(' ')[0].partition('_')[2]))[-1]
            matches['bottom copper'] = matches.pop(last_inner)

    return matches


def _layername_autoguesser(fn):
    fn, _, ext = fn.lower().rpartition('.')
    
    if ext in ('log', 'err', 'fdl', 'py', 'sh', 'md', 'rst', 'zip', 'pdf', 'svg', 'ps', 'png', 'jpg', 'bmp'):
        return 'unknown unknown'

    side, use = 'unknown', 'unknown'
    if re.search('top|front|pri?m?(ary)?', fn):
        side = 'top'
        use = 'copper'

    if re.search('bot(tom)?|back|sec(ondary)?', fn):
        side = 'bottom'
        use = 'copper'

    if re.search('silks?(creen)?|symbol', fn):
        use = 'silk'

    elif re.search('(solder)?paste|metalmask', fn):
        use = 'paste'

    elif re.search('(solder)?(mask|resist)', fn):
        use = 'mask'

    elif re.search('drill|rout?e?', fn):
        side = 'drill'
        use = 'unknown'

        if re.search(r'np(th|lt)?|(non|un)\W*plated|(non|un)\Wgalv', fn):
            use = 'nonplated'

        elif re.search('pth|plated|galv|plt', fn):
            use = 'plated'

    elif (m := re.search(r'(la?y?e?r?|in(ner)?|conduct(or|ive)?)\W*(?P<num>[0-9]+)', fn)):
        use = 'copper'
        side = f'inner_{int(m["num"]):02d}'

    elif re.search('film', fn):
        use = 'copper'

    elif re.search('out(line)?', fn):
        use = 'outline'
        side = 'mechanical'

    elif 'ipc' in fn and '356' in fn:
        use = 'netlist'
        side = 'other'

    elif 'netlist' in fn:
        use = 'netlist'
        side = 'other'

    if side == 'unknown':
        if re.search(r'[^a-z0-9]a', fn):
            side = 'top'
        elif re.search(r'[^a-z0-9]b', fn):
            side = 'bottom'

    return f'{side} {use}'


def _sort_layername(val):
    (side, use), _layer = val
    if side == 'top':
        return -1
    if side == 'bottom':
        return 1e99
    assert side.startswith('inner_')
    return int(side[len('inner_'):])


class LayerStack:
    """ :py:class:`LayerStack` represents a set of Gerber files that describe different layers of the same board.

    :ivar graphic_layers: :py:obj:`dict` mapping :py:obj:`(side, use)` tuples to the Gerber layers of the board.
                          :py:obj:`side` can be one of :py:obj:`"top"`, :py:obj:`"bottom"`, :py:obj:`"mechanical"`, or a
                          numbered internal layer such as :py:obj:`"inner2"`. :py:obj:`use` can be one of
                          :py:obj:`"silk", :py:obj:`mask`, :py:obj:`paste` or :py:obj:`copper`. For internal layers,
                          only :py:obj:`copper` is valid.
    :ivar board_name: Name of this board as parse from the input filenames, as a :py:obj:`str`. You can overwrite this
                      attribute with a different name, which will then be used during saving with the built-in file
                      naming rules.
    :ivar netlist: The :py:class:`~.ipc356.Netlist` of this board, or :py:obj:`None`
    :ivar original_path: The path to the directory or zip file that this board was loaded from.
    :ivar was_zipped: True if this board was loaded from a zip file.
    :ivar generator: A string containing an educated guess on which EDA tool generated this file. Example:
                     :py:obj:`"altium"`
    """

    def __init__(self, graphic_layers=None, drill_pth=None, drill_npth=None, drill_layers=(), netlist=None,
                 board_name=None, original_path=None, was_zipped=False, generator=None, courtyard=False,
                 fabrication=False, adhesive=False):
        if not drill_layers and (graphic_layers, drill_pth, drill_npth) == (None, None, None):
            graphic_layers = {tuple(layer.split()): GerberFile()
                    for layer in ('top paste', 'top silk', 'top mask', 'top copper',
                                  'bottom copper', 'bottom mask', 'bottom silk', 'bottom paste',
                                  'mechanical outline')}

            if courtyard:
                graphic_layers = {('top', 'courtyard'): GerberFile(),
                                  **graphic_layers,
                                  ('bottom', 'courtyard'): GerberFile()}

            if fabrication:
                graphic_layers = {('top', 'fabrication'): GerberFile(),
                                  **graphic_layers,
                                  ('bottom', 'fabrication'): GerberFile()}

            if adhesive:
                graphic_layers = {('top', 'adhesive'): GerberFile(),
                                  **graphic_layers,
                                  ('bottom', 'adhesive'): GerberFile()}

            drill_pth = ExcellonFile()
            drill_npth = ExcellonFile()

        self.graphic_layers = graphic_layers
        self.drill_pth = drill_pth
        self.drill_npth = drill_npth
        self._drill_layers = list(drill_layers)
        self.drill_mixed = None
        self.board_name = board_name
        self.netlist = netlist
        self.original_path = original_path
        self.was_zipped = was_zipped
        self.generator = generator

    @classmethod
    def open(kls, path, board_name=None, lazy=False, overrides=None, autoguess=True):
        """ Load a board from the given path.

        * The path can be a single file, in which case a :py:class:`LayerStack` containing only that file on a custom
          layer is returned.
        * The path can point to a directory, in which case the content's of that directory are analyzed for their file
          type and function.
        * The path can point to a zip file, in which case that zip file's contents are analyzed for their file type and
          function.
        * Finally, the path can be the string :py:obj:`"-"`, in which case this function will attempt to read a zip file
          from standard input.

        :param path: Path to a gerber file, directory or zip file, or the string :py:obj:`"-"`
        :param board_name: Override board name for the returned :py:class:`LayerStack` instance instead of guessing the
                           board name from the found file names.
        :param lazy: Do not parse files right away, instead return a :py:class:`LayerStack` containing
                     :py:class:~.cam.LazyCamFile` instances.
        :param overrides: :py:obj:`dict` containing a filename regex to layer type mapping that will override
                          gerbonara's built-in automatic rules. Each key must be a :py:obj:`str` containing a regex, and
                          each value must be a :py:obj:`(side, use)` :py:obj:`tuple` of :py:obj:`str`.
        :param autoguess: :py:obj:`bool` to enable or disable gerbonara's built-in automatic filename-based layer
                          function guessing. When :py:obj:`False`, layer functions are deduced only from
                          :py:obj:`overrides`.
        :rtype: :py:class:`LayerStack`
        """
        if str(path) == '-':
            data_io = io.BytesIO(sys.stdin.buffer.read())
            return kls.from_zip_data(data_io, original_path='<stdin>', board_name=board_name, lazy=lazy)

        path = Path(path)
        if path.is_dir():
            return kls.open_dir(path, board_name=board_name, lazy=lazy, overrides=overrides, autoguess=autoguess)
        elif path.suffix.lower() == '.zip' or is_zipfile(path):
            return kls.open_zip(path, board_name=board_name, lazy=lazy, overrides=overrides, autoguess=autoguess)
        else:
            return kls.from_files([path], board_name=board_name, lazy=lazy, overrides=overrides, autoguess=False)

    @classmethod
    def open_zip(kls, file, original_path=None, board_name=None, lazy=False, overrides=None, autoguess=True):
        """ Load a board from a ZIP file. Refer to :py:meth:`~.layers.LayerStack.open` for the meaning of the other
        options. 

        :param file: file-like object
        :param original_path: Override the :py:obj:`original_path` of the resulting :py:class:`LayerStack` with the
                              given value.
        :rtype: :py:class:`LayerStack`
        """
        tmpdir = tempfile.TemporaryDirectory()
        tmp_indir = Path(tmpdir.name) / 'input'
        tmp_indir.mkdir()

        with ZipFile(file) as f:
            f.extractall(path=tmp_indir)

        inst = kls.open_dir(tmp_indir, board_name=board_name, lazy=lazy)
        inst.tmpdir = tmpdir
        inst.original_path = Path(original_path or file)
        inst.was_zipped = True
        return inst

    @classmethod
    def open_dir(kls, directory, board_name=None, lazy=False, overrides=None, autoguess=True):
        """ Load a board from a directory. Refer to :py:meth:`~.layers.LayerStack.open` for the meaning of the options. 

        :param directory: Path of the directory to process.
        :rtype: :py:class:`LayerStack`
        """

        directory = Path(directory)
        if not directory.is_dir():
            raise FileNotFoundError(f'{directory} is not a directory')

        files = [ path for path in directory.glob('**/*') if path.is_file() ]
        return kls.from_files(files, board_name=board_name, lazy=lazy, original_path=directory, overrides=overrides,
                              autoguess=autoguess)
        inst.original_path = directory
        return inst

    @classmethod
    def from_files(kls, files, board_name=None, lazy=False, original_path=None, was_zipped=False, overrides=None,
                   autoguess=True):
        """ Load a board from a directory. Refer to :py:meth:`~.layers.LayerStack.open` for the meaning of the options. 

        :param files: List of paths of the files to load.
        :param original_path: Override the :py:obj:`original_path` of the resulting :py:class:`LayerStack` with the
                              given value.
        :param was_zipped: Override the :py:obj:`was_zipped` attribute of the resulting :py:class:`LayerStack` with the
                           given value.
        :rtype: :py:class:`LayerStack`
        """

        if autoguess:
            generator, filemap = _best_match(files)
        else:
            generator = 'custom'
            if overrides:
                filemap = {}
            else:
                filemap = {'unknown unknown': files}
        all_generator_hints = set()

        if overrides:
            for fn in files:
                for expr, layer in overrides.items():
                    if re.fullmatch(expr, fn.name):
                        if layer == 'ignore':
                            for entries in filemap.values():
                                if fn in entries:
                                    entries.remove(fn)
                        else:
                            if layer in filemap and fn in filemap[layer]:
                                filemap[layer].remove(fn)
                            filemap[layer] = filemap.get(layer, []) + [fn]

        if sum(len(files) for files in filemap.values()) < 6 and autoguess:
            warnings.warn('Ambiguous gerber filenames. Trying last-resort autoguesser.')
            generator = None
            filemap = _do_autoguess(files)
            if len(filemap) < 6:
                raise ValueError('Cannot figure out gerber file mapping. Partial map is: ', filemap)

        excellon_settings, external_tools = None, None
        if generator == 'geda':
            # geda is written by geniuses who waste no bytes of unnecessary output so it doesn't actually include the
            # number format in files that use imperial units. Unfortunately it also doesn't include any hints that the
            # file was generated by geda, so we have to guess by context whether this is just geda being geda or
            # potential user error.
            excellon_settings = FileSettings(number_format=(2, 4))

        elif generator == 'allegro':
            # Allegro puts information that is absolutely vital for parsing its excellon files... into another file,
            # next to the actual excellon file. Despite pretty much everyone else having figured out a way to put that
            # info into the excellon file itself, even if only as a comment.
            if 'excellon params' in filemap:
                excellon_settings = parse_allegro_ncparam(filemap['excellon params'][0].read_text())
                for file in filemap['excellon params']:
                    if (external_tools := parse_allegro_logfile(file.read_text())):
                        break
                del filemap['excellon params']
            # Ignore if we can't find the param file -- maybe the user has convinced Allegro to actually put this
            # information into a comment, or maybe they have made Allegro just use decimal points like XNC does.

            filemap = _do_autoguess([ f for files in filemap.values() for f in files ])
            if len(filemap) < 6:
                raise SystemError('Cannot figure out gerber file mapping')
            # FIXME use layer metadata from comments and ipc file if available

        elif generator == 'zuken':
            filemap = _do_autoguess([ f for files in filemap.values() for f in files ])
            if len(filemap) < 6:
                raise SystemError('Cannot figure out gerber file mapping')
            # FIXME use layer metadata from comments and ipc file if available

        elif generator == 'altium':
            if 'mechanical outline' in filemap:
                # Use lowest-numbered mechanical layer as outline, ignore others.
                mechs = {}
                for layer in filemap['mechanical outline']:
                    if layer.name.lower().endswith('gko'):
                        filemap['mechanical outline'] = [layer]
                        break

                    if (m := re.match(r'.*\.gm([0-9]+)', layer.name, re.IGNORECASE)):
                        mechs[int(m[1])] = layer
                    else:
                        break
                else:
                    filemap['mechanical outline'] = [sorted(mechs.items(), key=lambda x: x[0])[0][1]]

        else:
            excellon_settings = None

        ambiguous = [ f'{key} ({", ".join(x.name for x in value)})' for key, value in filemap.items() if len(value) > 1 and not 'drill' in key ]
        if ambiguous:
            raise SystemError(f'Ambiguous layer names: {", ".join(ambiguous)}')

        drill_pth, drill_npth = None, None
        drill_layers = []
        netlist = None
        layers = {} # { tuple(key.split()): None for key in STANDARD_LAYERS }
        for key, paths in filemap.items():
            if len(paths) > 1 and not 'drill' in key:
                raise ValueError(f'Multiple matching files found for {key} layer: {", ".join(value)}')

            for path in paths:
                id_result = identify_file(path.read_text())

                if 'netlist' in key:
                    layer = LazyCamFile(Netlist, path)

                elif ('outline' in key or 'drill' in key) and id_result != 'gerber':
                    if id_result is None:
                        # Since e.g. altium uses ".txt" as the extension for its drill files, we have to assume the
                        # current file might not be a drill file after all.
                        continue

                    if 'nonplated' in key:
                        plated = False
                    elif 'plated' in key:
                        plated = True
                    else:
                        plated = None
                    layer = LazyCamFile(ExcellonFile, path, plated=plated, settings=excellon_settings, external_tools=external_tools)
                else:

                    layer = LazyCamFile(GerberFile, path)

                if not lazy:
                    layer = layer.instance

                if key == 'mechanical outline':
                    layers['mechanical', 'outline'] = layer

                elif 'drill' in key:
                    if 'nonplated' in key and drill_npth is None:
                        drill_npth = layer
                    elif 'plated' in key and drill_pth is None:
                        drill_pth = layer
                    else:
                        drill_layers.append(layer)

                elif 'netlist' in key:
                    if netlist:
                        warnings.warn(f'Found multiple netlist files, using only first one. Have: {netlist.original_path.name}, got {path.name}')
                    else:
                        netlist = layer

                else:
                    side, _, use = key.partition(' ')
                    layers[(side, use)] = layer

                if not lazy:
                    hints = set(layer.generator_hints) | { generator }
                    all_generator_hints |= hints
                    if len(hints) > 1:
                        warnings.warn('File identification returned ambiguous results. Please raise an issue on the '
                                'gerbonara tracker and if possible please provide these input files for reference.')

        if not board_name:
            board_name = _common_prefix([l.original_path.name for l in layers.values() if l is not None])
            board_name = re.sub(r'^\W+', '', board_name)
            board_name = re.sub(r'\W+$', '', board_name)

        return kls(layers, drill_pth, drill_npth, drill_layers, board_name=board_name,
                original_path=original_path, was_zipped=was_zipped, generator=[*all_generator_hints, None][0])

    def save_to_zipfile(self, path, prefix='', overwrite_existing=True, board_name=None, naming_scheme={},
                          gerber_settings=None, excellon_settings=None):
        """ Save this board into a zip file at the given path. For other options, see
        :py:meth:`~.layers.LayerStack.save_to_directory`.

        :param path: Path of output zip file
        :param overwrite_existing: Bool specifying whether override an existing zip file. If :py:obj:`False` and
                                   :py:obj:`path` exists, a :py:obj:`ValueError` is raised.
        :param board_name: Board name to use when naming the Gerber/Excellon files

        :param prefix: Store output files under the given prefix inside the zip file
        """
        if path.is_file():
            if overwrite_existing:
                path.unlink()
            else:
                raise ValueError('output zip file already exists and overwrite_existing is False')

        if gerber_settings and not excellon_settings:
            excellon_settings = gerber_settings

        with ZipFile(path, 'w') as le_zip:
            for path, layer in self._save_files_iter(board_name=board_name, naming_scheme=naming_scheme):
                with le_zip.open(prefix + str(path), 'w') as out:
                    out.write(layer.instance.write_to_bytes())

    def save_to_directory(self, path, naming_scheme={}, overwrite_existing=True, board_name=None,
                          gerber_settings=None, excellon_settings=None):
        """ Save this board into a directory at the given path. If the given path does not exist, a new directory is
        created in its place.

        :param path: Output directory
        :param naming_scheme: :py:obj:`dict` specifying the naming scheme to use for the individual layer files. When
                              not specified, the original filenames are kept where available, and a default naming
                              scheme is used. You can provide your own :py:obj:`dict` here, mapping :py:obj:`"side use"`
                              strings to filenames, or use one of :py:attr:`~.layers.NamingScheme.kicad` or
                              :py:attr:`~.layers.NamingScheme.kicad`.
        :param board_name: Board name to use when naming the Gerber/Excellon files
        :param overwrite_existing: Bool specifying whether override an existing directory. If :py:obj:`False` and
                                   :py:obj:`path` exists, a :py:obj:`ValueError` is raised. Note that a
                                   :py:obj:`ValueError` will still be raised if the target exists and is not a
                                   directory.
        :param gerber_settings: :py:class:`~.cam.FileSettings` to use for Gerber file export. When not given, the input
                                file's original settings are re-used if available. If those can't be found anymore, sane
                                defaults are used. We recommend you set this to the result of
                                :py:meth:`~.cam.FileSettings.defaults`.
        """
        outdir = Path(path)
        outdir.mkdir(parents=True, exist_ok=overwrite_existing)

        if gerber_settings and not excellon_settings:
            excellon_settings = gerber_settings

        for path, layer in self._save_files_iter(board_name=board_name, naming_scheme=naming_scheme):
            out = outdir / path
            if out.exists() and not overwrite_existing:
                raise SystemError(f'Path exists but overwrite_existing is False: {out}')
            layer.instance.save(out)

    def _save_files_iter(self, board_name=None, naming_scheme={}):
        board_name = board_name or self.board_name

        if board_name is None:
            import inspect
            frame = inspect.currentframe()
            if frame is None:
                board_name = 'board'
            else:
                while frame is not None:
                    import sys
                    if not frame.f_globals['__name__'].startswith('gerbonara'):
                        board_name = frame.f_code.co_name
                        del frame
                        break
                    old_frame, frame = frame, frame.f_back
                    del old_frame

        def get_name(layer_type, layer):
            nonlocal naming_scheme, board_name

            if (m := re.match('inner_([0-9]+) copper', layer_type)):
                layer_type = 'inner copper'
                num = int(m[1])
            else:
                num = None

            if layer_type in naming_scheme:
                path = naming_scheme[layer_type].format(layer_number=num, board_name=board_name)
            elif layer.original_path and layer.original_path.name:
                path = layer.original_path.name
            else:
                path = NamingScheme.kicad[layer_type].format(layer_number=num, board_name=board_name)
                #ext = 'drl' if isinstance(layer, ExcellonFile) else 'gbr'
                #path = f'{board_name}-{layer_type.replace(" ", "_")}.{ext}'

            return path

        for (side, use), layer in self.graphic_layers.items():
            yield get_name(f'{side} {use}', layer), layer

        #self.normalize_drill_layers()

        if self.drill_pth is not None:
            yield get_name('drill plated', self.drill_pth), self.drill_pth

        if self.drill_npth is not None:
            yield get_name('drill nonplated', self.drill_npth), self.drill_npth

        for layer in self._drill_layers:
            yield get_name('drill unknown', layer), layer

        if self.netlist:
            yield get_name('other netlist', self.netlist), self.netlist

    def __str__(self):
        names = [ f'{side} {use}' for side, use in self.graphic_layers ]
        num_drill_layers = len(list(self.drill_layers))
        return f'<LayerStack {self.board_name} [{", ".join(names)}] and {num_drill_layers} drill layers>'

    def __repr__(self):
        return str(self)

    def to_svg(self, margin=0, side_re='.*', drills=True, arg_unit=MM, svg_unit=MM, force_bounds=None, colors=None, tag=Tag):
        """ Convert this layer stack to a plain SVG string. This is intended for use cases where the resulting SVG will
        be processed by other tools, and thus styling with colors or extra markup like Inkscape layer information are
        unwanted. If you want to instead generate a nice-looking preview image for display or graphical editing in tools
        such as Inkscape, use :py:meth:`~.layers.LayerStack.to_pretty_svg` instead.

        WARNING: The SVG files generated by this function preserve the Gerber coordinates 1:1, so the file will be
        mirrored vertically.

        :param margin: Export SVG file with given margin around the board's bounding box.
        :param side_re: A regex, such as ``'top'``, ``'bottom'``, or ``'.*'`` (default). Selects which layers to export.
                        The default includes inner layers.
        :param drills: :py:obj:`bool` setting if drills are included (default) or not. 
        :param arg_unit: :py:class:`.LengthUnit` or str (``'mm'`` or ``'inch'``). Which unit ``margin`` and
                         ``force_bounds`` are specified in. Default: mm
        :param svg_unit: :py:class:`.LengthUnit` or str (``'mm'`` or ``'inch'``). Which unit to use inside the SVG file.
                         Default: mm
        :param force_bounds: Use bounds given as :py:obj:`((min_x, min_y), (max_x, max_y))` tuple for the output SVG
                             file instead of deriving them from this board's bounding box and ``margin``. Note that this
                             will not scale or move the board, but instead will only crop the viewport.
        :param colors: Dict mapping ``f'{side} {use}'`` strings to SVG colors.
        :param tag: Extension point to support alternative XML serializers in addition to the built-in one.
        :rtype: :py:obj:`str`
        """
        if force_bounds:
            bounds = svg_unit.convert_bounds_from(arg_unit, force_bounds)
        else:
            bounds = self.bounding_box(svg_unit, default=((0, 0), (0, 0)))

        stroke_attrs = {'stroke_linejoin': 'round', 'stroke_linecap': 'round'}
        
        if colors is None:
            colors = defaultdict(lambda: 'black')
        
        tags = []
        layer_transform = f'translate(0 {bounds[0][1] + bounds[1][1]}) scale(1 -1)'
        for (side, use), layer in reversed(self.graphic_layers.items()):
            if re.fullmatch(side_re, side) and (fg := colors.get(f'{side} {use}')):
                tags.append(tag('g', list(layer.svg_objects(svg_unit=svg_unit, fg=fg, bg="white", tag=Tag)),
                        **stroke_attrs, id=f'l-{side}-{use}', transform=layer_transform))

        if drills:
            if self.drill_pth and (fg := colors.get('drill pth')):
                tags.append(tag('g', list(self.drill_pth.svg_objects(svg_unit=svg_unit, fg=fg, bg="white", tag=Tag)),
                        **stroke_attrs, id=f'l-drill-pth', transform=layer_transform))

            if self.drill_npth and (fg := colors.get('drill npth')):
                tags.append(tag('g', list(self.drill_npth.svg_objects(svg_unit=svg_unit, fg=fg, bg="white", tag=Tag)),
                        **stroke_attrs, id=f'l-drill-npth', transform=layer_transform))

            if (fg := colors.get('drill unknown')):
                for i, layer in enumerate(self._drill_layers):
                    tags.append(tag('g', list(layer.svg_objects(svg_unit=svg_unit, fg=fg, bg="white", tag=Tag)),
                            **stroke_attrs, id=f'l-drill-{i}', transform=layer_transform))

        return setup_svg(tags, bounds, margin=margin, arg_unit=arg_unit, svg_unit=svg_unit, tag=tag)

    def to_pretty_svg(self, side='top', margin=0, arg_unit=MM, svg_unit=MM, force_bounds=None, tag=Tag, inkscape=False,
                      colors=None, use=True):
        """ Convert this layer stack to a pretty SVG string that is suitable for display or for editing in tools such as
        Inkscape. If you want to process the resulting SVG in other tools, consider using
        :py:meth:`~layers.LayerStack.to_svg` instead, which produces output without color styling or blending based on
        SVG filter effects.

        :param side: One of the strings :py:obj:`"top"` or :py:obj:`"bottom"` specifying which side of the board to
                     render.
        :param margin: Export SVG file with given margin around the board's bounding box.
        :param arg_unit: :py:class:`.LengthUnit` or str (``'mm'`` or ``'inch'``). Which unit ``margin`` and
                         ``force_bounds`` are specified in. Default: mm
        :param svg_unit: :py:class:`.LengthUnit` or str (``'mm'`` or ``'inch'``). Which unit to use inside the SVG file.
                         Default: mm
        :param force_bounds: Use bounds given as :py:obj:`((min_x, min_y), (max_x, max_y))` tuple for the output SVG
                             file instead of deriving them from this board's bounding box and ``margin``. Note that this
                             will not scale or move the board, but instead will only crop the viewport.
        :param tag: Extension point to support alternative XML serializers in addition to the built-in one.
        :param inkscape: :py:obj:`bool` enabling Inkscape-specific markup such as Inkscape-native layers
        :param colors: Colorscheme to use, or :py:obj:`None` for the built-in pseudo-realistic green solder mask default
                       color scheme. When given, must be a dict mapping semantic :py:obj:`"side use"` layer names such
                       as :py:obj:`"top copper"` to a HTML-like hex color code such as :py:obj:`#ff00ea`. Transparency
                       is supported through 8-digit color codes. When 8 digits are given, the last two digits are used
                       as the layer's alpha channel. Valid side values in the layer name strings are :py:obj:`"top"`,
                       :py:obj:`"bottom"`, and :py:obj:`"mechanical"` as well as :py:obj:`"inner1"`, :py:obj:`"inner2"`
                       etc. for internal layers. Valid use values are :py:obj:`"mask"`, :py:obj:`"silk"`,
                       :py:obj:`"paste"`, and :py:obj:`"copper"`. For internal layers, only :py:obj:`"copper"` is valid.
        :param use: Enable/disable ``<use>`` tags for aperture flashes. Defaults to :py:obj:`True` (enabled).
        :rtype: :py:obj:`str`
        """
        if colors is None:
            colors = DEFAULT_COLORS
        use_use = use

        colors_alpha = {}
        for layer, color in colors.items():
            if isinstance(color, str):
                if re.match(r'#[0-9a-fA-F]{8}', color):
                    colors_alpha[layer] = (color[:-2], int(color[-2:], 16)/255)
                else:
                    colors_alpha[layer] = (color, 1)
            else:
                colors_alpha[layer] = color

        if force_bounds:
            bounds = svg_unit.convert_bounds_from(arg_unit, force_bounds)
        else:
            bounds = self.board_bounds(unit=svg_unit, default=((0, 0), (0, 0)))
        
        filter_defs = []

        for layer, (color, alpha) in colors_alpha.items():
            filter_defs.append(textwrap.dedent(f'''
                <filter id="f-{layer}">
                <feFlood result="flood-black" flood-color="black" flood-opacity="1"/>
                <feFlood result="flood-green" flood-color="{color}"/>
                <feBlend in="SourceGraphic" in2="flood-black" result="overlay" mode="normal"/>
                <feBlend in="overlay" in2="flood-green" result="colored" mode="multiply"/>
                <feColorMatrix in="overlay" type="matrix" result="alphaOut" values="0 0 0 0 0
                0 0 0 0 0
                0 0 0 0 0
                {alpha} 0 0 0 0"/>
                <feComposite in="colored" in2="alphaOut" operator="in"/>
                </filter>'''.strip()))

        inkscape_attrs = lambda label: dict(inkscape__groupmode='layer', inkscape__label=label) if inkscape else {}
        stroke_attrs = {'stroke_linejoin': 'round', 'stroke_linecap': 'round'}
        layer_transform=f'translate(0 {bounds[0][1] + bounds[1][1]}) scale(1 -1)'
        
        use_defs = []

        layers = []
        for use in ['copper', 'mask', 'silk', 'paste']:
            if (side, use) not in self:
                warnings.warn(f'Layer "{side} {use}" not found. Found layers: {", ".join(side + " " + use for side, use in self.graphic_layers)}')
                continue

            layer = self[(side, use)].instance

            fg, bg = ('white', 'black') if use != 'mask' else ('black', 'white')
            default_fill = {'copper': fg, 'mask': fg, 'silk': 'none', 'paste': fg}[use]
            default_stroke = {'copper': 'none', 'mask': 'none', 'silk': fg, 'paste': 'none'}[use]

            use_map = {}
            if use_use:
                layer.dedup_apertures()
                for obj in layer.objects:
                    if hasattr(obj, 'aperture') and obj.polarity_dark and obj.aperture not in use_map:
                        children = [prim.to_svg(fg, bg, tag=tag)
                                    for prim in obj.aperture.flash(0, 0, svg_unit, polarity_dark=True)]
                        use_id = f'a{len(use_defs)}'
                        use_defs.append(tag('g', children, id=use_id))
                        use_map[obj.aperture] = use_id

            objects = []
            for obj in layer.instance.svg_objects(svg_unit=svg_unit, fg=fg, bg=bg, aperture_map=use_map, tag=Tag):
                if obj.attrs.get('fill') == default_fill:
                    del obj.attrs['fill']
                elif 'fill' not in obj.attrs:
                    obj.attrs['fill'] = 'none'

                if obj.attrs.get('stroke') == default_stroke:
                    del obj.attrs['stroke']
                elif default_stroke != 'none' and 'stroke' not in obj.attrs:
                    obj.attrs['stroke'] = 'none'
                objects.append(obj)

            if use == 'mask':
                objects.insert(0, tag('path', id='outline-path', d=self.outline_svg_d(unit=svg_unit), fill='white'))
            layers.append(tag('g', objects, id=f'l-{side}-{use}', filter=f'url(#f-{use})',
                              fill=default_fill, stroke=default_stroke, **stroke_attrs,
                              **inkscape_attrs(f'{side} {use}'), transform=layer_transform))

        for i, layer in enumerate(self.drill_layers):
            layers.append(tag('g', list(layer.instance.svg_objects(svg_unit=svg_unit, fg='white', bg='black', tag=Tag)),
                id=f'l-drill-{i}', filter=f'url(#f-drill)', **stroke_attrs, **inkscape_attrs(f'drill-{i}'),
                transform=layer_transform))

        if self.outline:
            layers.append(tag('g', list(self.outline.instance.svg_objects(svg_unit=svg_unit, fg='white', bg='black', tag=Tag)),
                id=f'l-mechanical-outline', **stroke_attrs, **inkscape_attrs(f'outline'),
                transform=layer_transform))

        tags = [tag('defs', filter_defs + use_defs), *layers]
        return setup_svg(tags, bounds, margin=margin, arg_unit=arg_unit, svg_unit=svg_unit, pagecolor="white", tag=tag, inkscape=inkscape)

    def bounding_box(self, unit=MM, default=None):
        """ Calculate and return the bounding box of this layer stack. This bounding box will include all graphical
        objects on all layers and drill files. Consider using :py:meth:`~.layers.LayerStack.board_bounds` instead if you
        are interested in the actual board's bounding box, which usually will be smaller since there could be graphical
        objects sticking out of the board's outline, especially on drawing or silkscreen layers.

        :param unit: :py:class:`.LengthUnit` or str (``'mm'`` or ``'inch'``). Which unit to return results in. Default: mm
        :param default: Default value to return if there are no objects on any layer.
        :returns: ``((x_min, y_min), (x_max, y_max))`` tuple of floats.
        :rtype: tuple
        """
        return sum_bounds(( layer.bounding_box(unit, default=default)
            for layer in itertools.chain(self.graphic_layers.values(), self.drill_layers) ), default=default)

    def board_bounds(self, unit=MM, default=None):
        """ Calculate and return the bounding box of this board's outline. If this board has no outline, this function
        falls back to :py:meth:`~.layers.LayerStack.bounding_box`, returning the bounding box of all objects on all
        layers and drill files instead.

        :param unit: :py:class:`.LengthUnit` or str (``'mm'`` or ``'inch'``). Which unit to return results in. Default: mm
        :param default: Default value to return if there are no objects on any layer.
        :returns: ``((x_min, y_min), (x_max, y_max))`` tuple of floats.
        :rtype: tuple
        """
        if self.outline:
            return self.outline.instance.bounding_box(unit=unit, default=default)
        else:
            return self.bounding_box(unit=unit, default=default)

    def offset(self, x=0, y=0, unit=MM):
        """ Move all objects on all layers and drill files by the given amount in X and Y direction.

        :param x: :py:obj:`float` with length to move objects along X axis.
        :param y: :py:obj:`float` with length to move objects along Y axis.
        :param unit: :py:class:`.LengthUnit` or str (``'mm'`` or ``'inch'``). Which unit ``x`` and ``y`` are specified
                     in. Default: mm
        """
        for layer in itertools.chain(self.graphic_layers.values(), self.drill_layers):
            layer.offset(x, y, unit=unit)

    def rotate(self, angle, cx=0, cy=0, unit=MM):
        """ Rotate all objects on all layers and drill files by the given angle around the given center of rotation
        (default: coordinate origin (0, 0)).

        :param angle: Rotation angle in radians.
        :param cx: :py:obj:`float` with X coordinate of center of rotation. Default: :py:obj:`0`.
        :param cy: :py:obj:`float` with Y coordinate of center of rotation. Default: :py:obj:`0`.
        :param unit: :py:class:`.LengthUnit` or str (``'mm'`` or ``'inch'``). Which unit ``cx`` and ``cy`` are specified
                     in. Default: mm
        """
        for layer in itertools.chain(self.graphic_layers.values(), self.drill_layers):
            layer.rotate(angle, cx, cy, unit=unit)

    def scale(self, factor, unit=MM):
        """ Scale all objects on all layers and drill files by the given scaling factor. Only uniform scaling with one
        common factor for both X and Y is supported since non-uniform scaling would not work with either arcs or
        apertures in Gerber or Excellon files.

        :param factor: Scale factor. :py:obj:`1.0` for no scaling, :py:obj:`2.0` for doubling in both directions.
        :param unit: :py:class:`.LengthUnit` or str (``'mm'`` or ``'inch'``) for compatibility with other transform
                     methods. Default: mm
        """

        for layer in itertools.chain(self.graphic_layers.values(), self.drill_layers):
            layer.scale(factor)

    def merge_drill_layers(self):
        """ Merge all drill layers of this board into a single drill layer containing all objects. You can access this
        drill layer under the :py:attr:`.LayerStack.drill_mixed` attribute. The original layers are removed from the
        board. """
        target = ExcellonFile(comments=['Drill files merged by gerbonara'])

        for layer in self.drill_layers:
            if isinstance(layer, GerberFile):
                layer = layer.to_excellon()

            target.merge(layer)

        self.drill_pth = self.drill_npth = None
        self.drill_mixed = target

    def normalize_drill_layers(self):
        """ Take everything from all drill layers of this board, and sort it into three new drill layers: One with all
        non-plated objects, one with all plated objects, and one for all leftover objects with unknown plating. This
        method replaces the board's drill layers with these three sorted ones. """
        # TODO: maybe also separate into drill and route?
        drill_pth, drill_npth, drill_aux = [], [], []

        for layer in self.drill_layers:
            if isinstance(layer, GerberFile):
                layer = layer.to_excellon()

            if layer.is_plated:
                drill_pth.append(layer)
            elif layer.is_nonplated:
                drill_pth.append(layer)
            else:
                drill_aux.append(layer)
        
        pth_out, *rest = drill_pth or [ExcellonFile()]
        for layer in rest:
            pth_out.merge(layer)

        npth_out, *rest = drill_npth or [ExcellonFile()]
        for layer in rest:
            npth_out.merge(layer)

        unknown_out = ExcellonFile()
        for layer in drill_aux:
            for obj in layer.objects:
                if obj.plated is None:
                    unknown_out.append(obj)
                elif obj.plated:
                    pth_out.append(obj)
                else:
                    npth_out.append(obj)

        self.drill_pth, self.drill_npth = pth_out, npth_out
        self._drill_layers = [unknown_out] if unknown_out else []

    @property
    def drill_layers(self):
        """ Generator iterating all of this board's drill layers. """
        if self.drill_pth:
            yield self.drill_pth

        if self.drill_npth:
            yield self.drill_npth

        if self._drill_layers:
            yield from self._drill_layers

    @drill_layers.setter
    def drill_layers(self, value):
        self._drill_layers = value
        self.drill_pth = self.drill_npth = None

    def __len__(self):
        return len(self.graphic_layers)

    def get(self, index, default=None):
        if index in self:
            return self[index]
        else:
            return default

    def __contains__(self, index):
        if isinstance(index, str):
            side, _, use = index.partition(' ')
            return (side, use) in self.graphic_layers

        elif isinstance(index, tuple):
            return index in self.graphic_layers

        return index < len(self.copper_layers)

    def __getitem__(self, index):
        if isinstance(index, str):
            side, _, use = index.partition(' ')
            return self.graphic_layers[(side, use)]

        elif isinstance(index, tuple):
            return self.graphic_layers[index]

        return self.copper_layers[index][1]

    def __setitem__(self, index, value):
        if isinstance(index, str):
            side, _, use = index.partition(' ')
            self.graphic_layers[(side, use)] = value

        elif isinstance(index, tuple):
            self.graphic_layers[index] = value

        else:
            raise IndexError('Layer {index} not found. Valid layer indices are "{side} {use}" strings or (side, use) tuples.')

    def add_layer(self, index):
        self[index] = GerberFile()

    @property
    def copper_layers(self):
        """ Return all copper layers of this board as a list of ((side, use), layer) tuples. Returns an empty list if
        the board does not have any copper layers. """
        layers = [((side, use), layer) for (side, use), layer in self.graphic_layers.items() if use == 'copper']
        return sorted(layers, key=_sort_layername)

    @property
    def inner_layers(self):
        """ Return all inner copper layers of this board as a list of ((side, use), layer) tuples. Returns an empty list
        if the board does not have any inner layers. """
        layers = [((side, use), layer) for (side, use), layer in self.graphic_layers.items() if side.startswith('inner')]
        return sorted(layers, key=_sort_layername)

    @property
    def top_side(self):
        """ Return a dict containing the subset of layers from :py:meth:`~.layers.LayerStack.graphic_layers` that are on
        the board's top side. Includes the board outline layer, if available. """
        return { key: self[key] for key in ('top copper', 'top mask', 'top silk', 'top paste', 'mechanical outline') }

    @property
    def bottom_side(self):
        """ Return a dict containing the subset of layers from :py:meth:`~.layers.LayerStack.graphic_layers` that are on
        the board's bottom side. Includes the board outline layer, if available. """
        return { key: self[key] for key in ('bottom copper', 'bottom mask', 'bottom silk', 'bottom paste', 'mechanical outline') }

    @property
    def outline(self):
        """ Return this board's outline layer if available, or :py:obj:`None`. """
        return self.get('mechanical outline')

    def outline_svg_d(self, tol=0.01, unit=MM):
        """ Return this board's outline as SVG path data.

        :param tol: :py:obj:`float` setting the tolerance below which two points are considered equal
        :param unit: :py:class:`.LengthUnit` or str (``'mm'`` or ``'inch'``). SVG document unit. Default: mm
        """
        chains = self.outline_polygons(tol, unit)
        polys = []
        for chain in chains:
            outline = [ (chain[0].x1, chain[0].y1), *((elem.x2, elem.y2) for elem in chain) ]
            arcs = [ (elem.clockwise, (elem.cx, elem.cy)) if isinstance(elem, gp.Arc) else None for elem in chain ]
            poly = gp.ArcPoly(outline=outline, arc_centers=arcs)
            polys.append(' '.join(poly.path_d()) + ' Z')
        return ' '.join(polys)

    def outline_convex_hull(self, tol=0.01, unit=MM):
        points = []
        for obj in self.outline.instance.objects:
            if isinstance(obj, go.Line):
                line = obj.as_primitive(unit)
                points.append((line.x1, line.y1))
                points.append((line.x2, line.y2))

            elif isinstance(obj, go.Arc):
                for obj in obj.approximate(tol, unit):
                    line = obj.as_primitive(unit)
                    points.append((line.x1, line.y1))
                    points.append((line.x2, line.y2))

        return convex_hull(points)
        
    def outline_polygons(self, tol=0.01, unit=MM):
        """ Iterator yielding this boards outline as a list of ordered :py:class:`~.graphic_objects.Arc` and
        :py:class:`~.graphic_objects.Line` objects. This method first sorts all lines and arcs on the outline layer into
        connected components, then orders them such that one object's end point is the next object's start point,
        flipping them where necessary. It yields one list of (likely mixed) :py:class:`~.graphic_objects.Arc` and
        :py:class:`~.graphic_objects.Line` objects per connected component.

        This method exists because the only convention in Gerber or Excellon outline files is that the outline segments
        are *visually contiguous*, but that does not necessarily mean that they will be in any particular order inside
        the G-code.

        :param tol: :py:obj:`float` setting the tolerance below which two points are considered equal
        :param unit: :py:class:`.LengthUnit` or str (``'mm'`` or ``'inch'``). SVG document unit. Default: mm
        """
        polygons = []
        lines = [ obj.as_primitive(unit) for obj in self.outline.instance.objects if isinstance(obj, (go.Line, go.Arc)) ]

        by_x = sorted([ (obj.x1, obj) for obj in lines ] + [ (obj.x2, obj) for obj in lines ], key=lambda x: x[0])
        dist_sq = lambda x1, y1, x2, y2: (x2-x1)**2 + (y2-y1)**2

        joins = {}
        for cur in lines:
            # Special case: An arc may describe a complete circle, in which case we have to return it as-is since it
            # is the only primitive that can join itself.
            if isinstance(cur, gp.Arc) and cur.is_circle:
                yield [cur]
                continue

            for (i, x, y) in [(0, cur.x1, cur.y1), (1, cur.x2, cur.y2)]:
                x_left  = bisect.bisect_left (by_x, x, key=lambda elem: elem[0] + tol)
                x_right = bisect.bisect_right(by_x, x, key=lambda elem: elem[0] - tol)
                selected = { elem for elem_x, elem in by_x[x_left:x_right] if elem != cur }

                if not selected:
                    continue # loose end

                nearest = sorted(selected, key=lambda elem: min(dist_sq(elem.x1, elem.y1, x, y), dist_sq(elem.x2, elem.y2, x, y)))[0]

                d1, d2 = dist_sq(nearest.x1, nearest.y1, x, y), dist_sq(nearest.x2, nearest.y2, x, y)
                j = 0 if d1 < d2 else 1

                if (nearest, j) in joins and joins[(nearest, j)] != (cur, i):
                    warnings.warn(f'Three-way intersection on outline layer at: {(nearest, j)}; {(cur, i)}; and {joins[(nearest, j)]}. Falling back to returning the convex hull of the outline layer.')
                    return self.outline_convex_hull(tol, unit)

                if (cur, i) in joins and joins[(cur, i)] != (nearest, j):
                    warnings.warn(f'three-way intersection on outline layer at: {(nearest, j)}; {(cur, i)}; and {joins[(nearest, j)]}. Falling back to returning the convex hull of the outline layer.')
                    return self.outline_convex_hull(tol, unit)


                joins[(cur, i)] = (nearest, j)
                joins[(nearest, j)] = (cur, i)

        def flip_if(obj, cond):
            if cond:
                return obj.flip()
            else:
                return obj

        while joins:
            (first, i), (cur, j) = joins.popitem()
            del joins[(cur, j)]
            l = [ flip_if(first, not i), flip_if(cur, j) ]
            while cur != first and (cur, not j) in joins:
                cur, j = joins.pop((cur, not j))
                del joins[(cur, j)]
                l.append(flip_if(cur, j))
            yield l


    def _merge_layer(self, target, source, mode='above'):
        if source is None:
            return
        
        if self[target] is None:
            self[target] = source

        else:
            self[target].merge(source, mode)

    def merge(self, other, mode='above'):
        """ Merge ``other`` into ``self``, i.e. for all layers, add all objects that are in ``other`` to ``self``. This
        resets :py:attr:`.import_settings` and :py:attr:`~.CamFile.generator` on all layers. Units and other
        file-specific settings are handled automatically. For the meaning of the ``mode`` parameter, see
        :py:meth:`.GerberFile.merge`.

        Layers are matched by their logical side and function as they are found in
        :py:meth:`.LayerStack.graphic_layers`. Drill layers are normalized before merging, which splits them into
        exactly three drill layers: An non-plated one, a plated one, and a (hopefully empty) unknown plating one.
        """
        all_keys = set(self.graphic_layers.keys()) | set(other.graphic_layers.keys())
        exclude = { tuple(key.split()) for key in STANDARD_LAYERS }
        all_keys = { key for key in all_keys if key not in exclude }
        if all_keys:
            warnings.warn('Cannot merge unknown layer types: {" ".join(all_keys)}')

        for side in 'top', 'bottom':
            for use in 'copper', 'mask', 'silk', 'paste':
                if (side, use) in other:
                    self._merge_layer((side, use), other[side, use], mode)

        our_inner, their_inner = self.copper_layers[1:-1], other.copper_layers[1:-1]

        if bool(our_inner) != bool(their_inner):
            warnings.warn('Merging board without inner layers into board with inner layers, inner layers will be empty on first board.')

        elif our_inner and their_inner:
            warnings.warn('Merging boards with different inner layer counts. Will fill inner layers starting at core.')

        diff = len(our_inner) - len(their_inner)
        their_inner = ([None] * max(0, diff//2)) + their_inner + ([None] * max(0, diff//2))
        our_inner = ([None] * max(0, -diff//2)) + their_inner + ([None] * max(0, -diff//2))

        new_inner = []
        for ours, theirs in zip(our_inner, their_inner):
            if ours is None:
                new_inner.append(theirs)
            elif theirs is None:
                new_inner.append(ours)
            else:
                ours.merge(theirs)
                new_inner.append(ours)

        for i, layer in enumerate(new_inner, start=1):
            self[f'inner_{i} copper'] = layer

        self._merge_layer('mechanical outline', other['mechanical outline'])

        self.normalize_drill_layers()
        other.normalize_drill_layers()

        self.drill_pth.merge(other.drill_pth)
        self.drill_npth.merge(other.drill_npth)
        self._drill_layers.extend(other._drill_layers)

        if self.netlist:
            self.netlist.merge(other.netlist)
        else:
            self.netlist = other.netlist




#!/usr/bin/env python3

import math

from gerbonara.graphic_objects import Arc
from gerbonara.graphic_objects import rotate_point


def approx_test():
    print('<?xml version="1.0" encoding="utf-8"?>')
    print('<svg width="300mm" height="300mm" viewBox="0 0 300 300" xmlns="http://www.w3.org/2000/svg">')
    n = 16
    eps = 1/n*2*math.pi
    cx, cy = 0, 0
    for clockwise in False, True:
        for start_angle in range(n):
            cx, cy = 0, cy+2.5
            for sweep_angle in range(n):
                for color, max_error in zip(['black', 'red', 'blue', 'magenta'], [0.1, 0.3, 1, 3]):
                    cx = cx+2.5

                    x1, y1 = rotate_point(0, -1, start_angle*eps)
                    x2, y2 = rotate_point(x1, y1, sweep_angle*eps*(-1 if clockwise else 1))
                    
                    arc = Arc(x1+cx, y1+cy, x2+cx, y2+cy, -x1, -y1, clockwise=clockwise, aperture=None, polarity_dark=True)
                    lines = arc.approximate(max_error=max_error)

                    print(f'<path style="fill: {color}; stroke: none;" d="M {cx} {cy} L {lines[0].x1} {lines[0].y1}', end=' ')
                    for line in lines:
                        print(f'L {line.x2} {line.y2}', end=' ')
                    print('"/>')
    print('</svg>')


if __name__ == '__main__':
    approx_test()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2014 Hamilton Kibbe <ham@hamiltonkib.be>
# Copyright 2022 Jan Sebastian Götte <gerbonara@jaseg.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import math
import operator
import warnings
import functools
import dataclasses
import re
from enum import Enum
from dataclasses import dataclass
from collections import Counter
from pathlib import Path

from .cam import CamFile, FileSettings
from .graphic_objects import Flash, Line, Arc
from .apertures import ExcellonTool
from .utils import Inch, MM, to_unit, InterpMode, RegexMatcher

class ExcellonContext:
    """ Internal helper class used for tracking graphics state when writing Excellon. """

    def __init__(self, settings, tools):
        self.settings = settings
        self.tools = tools
        self.mode = None
        self.current_tool = None
        self.x, self.y = None, None
        self.drill_down = False

    def select_tool(self, tool):
        """ Select the current tool. Retract drill first if necessary. """
        current_id = self.tools.get(self.current_tool)
        new_id = self.tools[tool]
        if new_id != current_id:
            if self.drill_down:
                yield 'M16' # drill up
                self.drill_down = False

            self.current_tool = tool
            yield f'T{new_id:02d}'

    def drill_mode(self):
        """ Enter drill mode. """
        if self.mode != ProgramState.DRILLING:
            self.mode = ProgramState.DRILLING
            if self.drill_down:
                yield 'M16' # drill up
                self.drill_down = False
            yield 'G05' # drill mode

    def route_mode(self, unit, x, y):
        """ Enter route mode and plunge tool at the given coordinates. """
        x, y = self.settings.unit(x, unit), self.settings.unit(y, unit)

        if self.mode == ProgramState.ROUTING and (self.x, self.y) == (x, y):
            return # nothing to do

        if self.drill_down:
            yield 'M16' # drill up

        # route mode
        yield 'G00' + 'X' + self.settings.write_excellon_value(x) + 'Y' + self.settings.write_excellon_value(y)
        yield 'M15' # drill down
        self.drill_down = True
        self.mode = ProgramState.ROUTING
        self.x, self.y = x, y

    def set_current_point(self, unit, x, y):
        """ Update internal last point """
        self.x, self.y = self.settings.unit(x, unit), self.settings.unit(y, unit)

def parse_allegro_ncparam(data, settings=None):
    """ Internal function to parse Excellon format information out of Allegro's nonstandard textual parameter files that
    it generates along with the Excellon file. """
    # This function parses data from allegro's nc_param.txt and ncdrill.log files. We have to parse these files because
    # allegro Excellon files omit crucial information such as the *number format*. nc_param.txt really is the file we
    # want to parse, but sometimes due to user error it doesn't end up in the gerber package. In this case, we want to
    # still be able to extract the same information from the human-readable ncdrill.log.

    if settings is None:
        settings = FileSettings(number_format=(None, None), zeros='leading')

    lz_supp, tz_supp = False, False
    nf_int, nf_frac = settings.number_format
    for line in data.splitlines():
        line = re.sub(r'\s+', ' ', line.strip())

        if (match := re.fullmatch(r'FORMAT ([0-9]+\.[0-9]+)', line)):
            x, _, y = match[1].partition('.')
            nf_int, nf_frac = int(x), int(y)

        elif (match := re.fullmatch(r'INTEGER-PLACES ([0-9]+)', line)):
            nf_int = int(match[1])

        elif (match := re.fullmatch(r'DECIMAL-PLACES ([0-9]+)', line)):
            nf_frac = int(match[1])

        elif (match := re.fullmatch(r'COORDINATES (ABSOLUTE|.*)', line)):
            # I have not been able to find a single incremental-notation allegro file. Probably that is for the better.
            settings.notation = match[1].lower()

        elif (match := re.fullmatch(r'OUTPUT-UNITS (METRIC|ENGLISH|INCHES)', line)):
            # I have no idea wth is the difference between "ENGLISH" and "INCHES". I think one might just be the one
            # Allegro uses in footprint files, with the other one being used in gerber exports.
            settings.unit = MM if match[1] == 'METRIC' else Inch

        elif (match := re.fullmatch(r'SUPPRESS-LEAD-ZEROES (YES|NO)', line)):
            lz_supp = (match[1] == 'YES')

        elif (match := re.fullmatch(r'SUPPRESS-TRAIL-ZEROES (YES|NO)', line)):
            tz_supp = (match[1] == 'YES')

    if lz_supp and tz_supp:
        raise SyntaxError('Allegro Excellon parameters specify both leading and trailing zero suppression. We do not '
                'know how to parse this. Please raise an issue on our issue tracker and provide an example file.')

    settings.number_format = nf_int, nf_frac
    settings.zeros = 'leading' if lz_supp else 'trailing'
    return settings


def parse_allegro_logfile(data):
    """ Internal function to parse Excellon format information out of Allegro's nonstandard textual log files that it
    generates along with the Excellon file. """
    found_tools = {}
    unit = None

    for line in data.splitlines():
        line = line.strip()
        line = re.sub(r'\s+', ' ', line)

        if (m := re.match(r'OUTPUT-UNITS (METRIC|ENGLISH|INCHES)', line)):
            # I have no idea wth is the difference between "ENGLISH" and "INCHES". I think one might just be the one
            # Allegro uses in footprint files, with the other one being used in gerber exports.
            unit = MM if m[1] == 'METRIC' else Inch

        elif (m := re.match(r'T(?P<index1>[0-9]+) (?P<index2>[0-9]+)\. (?P<diameter>[0-9/.]+) [0-9. /+-]* (?P<plated>PLATED|NON_PLATED|OPTIONAL) [0-9]+', line)):
            index1, index2 = int(m['index1']), int(m['index2'])
            if index1 != index2:
                return {}

            diameter = float(m['diameter'])
            if unit == Inch:
                diameter /= 1000
            is_plated = None if m['plated'] is None else (m['plated'] in ('PLATED', 'OPTIONAL'))
            found_tools[index1] = ExcellonTool(diameter=diameter, plated=is_plated, unit=unit)
    return found_tools

def parse_zuken_logfile(data):
    """ Internal function to parse Excellon format information out of Zuken's nonstandard textual log files that their
    tools generate along with the Excellon file. """
    lines = [ line.strip() for line in data.splitlines() ]
    if '*****  DRILL LIST  *****' not in lines:
        return # likely not a Zuken CR-8000 logfile 

    params = {}
    for line in lines:
        key, colon, value = line.partition(':')
        if colon and value:
            params[key.strip()] = value.strip()

    if not (fmt := params.get('Coordinate Format')):
        return None

    integer, _, decimal = fmt.partition('V')
    settings = FileSettings(number_format=(int(integer), int(decimal)))
    
    if (supp := params.get('Zero Suppress')):
        supp, _1, _2 = supp.partition(' ')
        settings.zeros = supp.lower()

    return settings


class ExcellonFile(CamFile):
    """ Excellon drill file.

    An Excellon file can contain both drills and milled slots. Drills are represented by :py:class:`.Flash` instances
    with their aperture set to the special :py:class:`.ExcellonDrill` aperture class. Drills can be plated or nonplated.
    This information is stored in the :py:class:`.ExcellonTool`. Both can co-exist in the same file, and some CAD tools
    even export files like this. :py:class:`.LayerStack` contains functions to convert between a single drill file with
    mixed plated and nonplated holes and one with separate drill files for each. Best practice is to have separate drill
    files for slots, nonplated holes, and plated holes, because the board house will produce all three in three separate
    processes anyway, and also because there is no standardized way to represent plating in Excellon files. Gerbonara
    uses Altium's convention for this, which uses a magic comment before the tool definition.
    """

    def __init__(self, objects=None, comments=None, import_settings=None, original_path=None, generator_hints=None):
        super().__init__(original_path=original_path)
        self.objects = objects or []
        self.comments = comments or []
        self.import_settings = import_settings
        self.generator_hints = generator_hints or [] # This is a purely informational goodie from the parser. Use it as you wish.

    def __str__(self):
        name = f'{self.original_path.name} ' if self.original_path else ''
        return f'<ExcellonFile {name}{self.plating_type} with {len(list(self.drills()))} drills, {len(list(self.slots()))} slots using {len(self.drill_sizes())} tools>'

    def __repr__(self):
        return str(self)

    @property
    def plating_type(self):
        if self.is_plated:
            return 'plated'
        elif self.is_nonplated:
            return 'nonplated'
        elif self.is_mixed_plating:
            return 'mixed plating'
        else:
            return 'unknown plating'

    @property
    def is_plated(self):
        """ Test if *all* holes or slots in this file are plated. """
        return all(obj.plated for obj in self.objects)

    @property
    def is_nonplated(self):
        """ Test if *all* holes or slots in this file are non-plated. """
        return all(obj.plated == False for obj in self.objects) # False, not None

    @property
    def is_plating_unknown(self):
        """ Test if *all* holes or slots in this file have no known plating. """
        return all(obj.plated is None for obj in self.objects) # False, not None

    @property
    def is_mixed_plating(self):
        """ Test if there are multiple plating values used in this file. """
        return len({obj.plated for obj in self.objects}) > 1

    @property
    def is_plated_tristate(self):
        if self.is_plated:
            return True

        if self.is_nonplated:
            return False

        return None

    def append(self, obj_or_comment):
        """ Add a :py:class:`.GraphicObject` or a comment (str) to this file. """
        if isinstance(obj_or_comment, str):
            self.comments.append(obj_or_comment)
        else:
            self.objects.append(obj_or_comment)

    def to_excellon(self, plated=None, errors='raise'):
        """ Counterpart to :py:meth:`~.rs274x.GerberFile.to_excellon`. Does nothing and returns :py:obj:`self`. """
        return self

    def to_gerber(self, errros='raise'):
        """ Convert this excellon file into a :py:class:`~.rs274x.GerberFile`. """
        out = GerberFile()
        out.comments = self.comments

        apertures = {}
        for obj in self.objects:
            if not (ap := apertures[obj.tool]):
                ap = apertures[obj.tool] = CircleAperture(obj.tool.diameter)

            out.objects.append(dataclasses.replace(obj, aperture=ap))

    @property
    def generator(self):
        return self.generator_hints[0] if self.generator_hints else None

    def merge(self, other, mode='ignored', keep_settings=False):
        if other is None:
            return
        
        if not isinstance(other, ExcellonFile):
            other = other.to_excellon(plated=self.is_plated_tristate)

        self.objects += other.objects
        self.comments += other.comments
        self.generator_hints = None
        if not keep_settings:
            self.import_settings = None

    @classmethod
    def open(kls, filename, plated=None, settings=None, external_tools=None):
        """ Load an Excellon file from the file system.

        Certain CAD tools do not put any information on decimal points into the actual excellon file, and instead put
        that information into a non-standard text file next to the excellon file. Using :py:meth:`~.ExcellonFile.open`
        to open a file gives Gerbonara the opportunity to try to find this data. In contrast to pcb-tools, Gerbonara
        will raise an exception instead of producing garbage parsing results if it cannot determine the file format
        parameters with certainty.

        .. note:: This is preferred over loading Excellon from a str through :py:meth:`~.ExcellonFile.from_string`.

        :param filename: ``str`` or ``pathlib.Path``.
        :param bool plated: If given, set plating status of any tools in this file that have undefined plating. This is
                useful if you already know that this file contains only e.g. plated holes from contextual information
                such as the file name.
        :param FileSettings settings: Format settings to use. If None, try to auto-detect file settings.
        """

        filename = Path(filename)
        external_tools = None
    
        if settings is None:
            # Parse allegro parameter files for settings.
            # Prefer nc_param.txt over ncparam.log since the txt is the machine-readable one.
            for fn in 'nc_param.txt', 'ncdrill.log':
                if (param_file := filename.parent / fn).is_file():
                    settings =  parse_allegro_ncparam(param_file.read_text())
                    warnings.warn(f'Loaded allegro-style excellon settings file {param_file}')
                    break

            # Parse Zuken log file for settings
            if filename.name.endswith('.fdr'):
                logfile = filename.with_suffix('.fdl')
                if logfile.is_file():
                    settings = parse_zuken_logfile(logfile.read_text())
                    warnings.warn(f'Loaded zuken-style excellon log file {logfile}: {settings}')

        if external_tools is None:
            # Parse allegro log files for tools.
            # TODO add try/except aronud this
            log_file = filename.parent / 'ncdrill.log'
            if log_file.is_file():
                external_tools = parse_allegro_logfile(log_file.read_text())


        return kls.from_string(filename.read_text(), settings=settings,
                filename=filename, plated=plated, external_tools=external_tools)

    @classmethod
    def from_string(kls, data, settings=None, filename=None, plated=None, external_tools=None):
        """ Parse the given string as an Excellon file. Note that often, Excellon files do not contain any information
        on which number format (integer/decimal places, zeros suppression) is used. In case Gerbonara cannot determine
        this with certainty, this function *will* error out. Use :py:meth:`~.ExcellonFile.open` if you want Gerbonara to
        parse this metadata from the non-standardized text files many CAD packages produce in addition to drill files.
        """

        parser = ExcellonParser(settings, external_tools=external_tools)
        parser.do_parse(data, filename=filename)
        return kls(objects=parser.objects, comments=parser.comments, import_settings=parser.settings,
                generator_hints=parser.generator_hints, original_path=filename)

    def _generate_statements(self, settings, drop_comments=True):
        """ Export this file as Excellon code, yields one str per line. """
        yield '; XNC file generated by gerbonara'
        if self.comments and not drop_comments:
            yield '; Comments found in original file:'
            for comment in self.comments:
                yield ';' + comment

        yield 'M48'
        yield 'METRIC' if settings.unit == MM else 'INCH'

        # Build tool index
        tool_map = { obj.tool: obj.tool for obj in self.objects }
        tools = sorted(tool_map.items(), key=lambda id_tool: (id_tool[1].plated, id_tool[1].diameter))

        mixed_plating = (len({ tool.plated for tool in tool_map.values() }) > 1)
        if mixed_plating:
            warnings.warn('Multiple plating values in same file. Will use non-standard Altium comment syntax to indicate hole plating.')

        defined_tools = {}
        tool_indices = {}
        index = 1
        for tool_id, tool in tools:
            xnc = tool.to_xnc(settings)
            if (tool.plated, xnc) in defined_tools:
                tool_indices[tool_id] = defined_tools[(tool.plated, xnc)]

            else:
                if mixed_plating:
                    yield ';TYPE=PLATED' if tool.plated else ';TYPE=NON_PLATED'

                yield f'T{index:02d}' + xnc

                tool_indices[tool_id] = defined_tools[(tool.plated, xnc)] = index
                index += 1

                if index >= 100:
                    warnings.warn('More than 99 tools defined. Some programs may not like three-digit tool indices.', SyntaxWarning)

        yield '%'

        ctx = ExcellonContext(settings, tool_indices)

        # Export objects
        for obj in self.objects:
            yield from obj.to_xnc(ctx)

        yield 'M30'

    def write_to_bytes(self, settings=None, drop_comments=True):
        """ Export to Excellon format. This function always generates XNC, which is a well-defined subset of Excellon.
        Uses sane default settings if you don't give any.


        :param bool drop_comments: If true, do not write comments to output file. This defaults to true because
                otherwise there is a risk that Gerbonara does not consider some obscure magic comment semantically
                meaningful while some other Excellon viewer might still parse it.
        
        :rtype: str
        """

        if settings is None:
            if self.import_settings:
                settings = self.import_settings.copy()
                settings.zeros = None
                settings.number_format = (3,5)
            else:
                settings = FileSettings.defaults()
        return '\n'.join(self._generate_statements(settings, drop_comments=drop_comments)).encode('utf-8')

    def save(self, filename, settings=None, drop_comments=True):
        """ Save this Excellon file to the file system. See :py:meth:`~.ExcellonFile.generate_excellon` for the meaning
        of the arguments. """
        with open(filename, 'wb') as f:
            f.write(self.write_to_bytes(settings, drop_comments=drop_comments))

    def offset(self, x=0, y=0, unit=MM):
        for obj in self.objects:
            obj.offset(x, y, unit)

    def rotate(self, angle, cx=0, cy=0, unit=MM):
        if math.isclose(angle % (2*math.pi), 0):
            return

        for obj in self.objects:
            obj.rotate(angle, cx, cy, unit=unit)

    def __len__(self):
        return len(self.objects)

    def split_by_plating(self):
        """ Split this file into two :py:class:`.ExcellonFile` instances, one containing all plated objects, and one
        containing all nonplated objects. In this function, objects with undefined plating are considered nonplated.

        .. note:: This does not copy the objects, so modifications in either of the returned files may clobber the
                  original file.

        :returns: (nonplated_file, plated_file)
        :rtype: tuple
        """
        plated = ExcellonFile(
            comments = self.comments.copy(),
            import_settings = self.import_settings.copy(),
            objects = [ obj for obj in self.objects if obj.plated ],
            filename = self.filename)

        nonplated = ExcellonFile(
            comments = self.comments.copy(),
            import_settings = self.import_settings.copy(),
            objects = [ obj for obj in self.objects if not obj.plated ],
            filename = self.filename)

        return nonplated, plated

    def path_lengths(self, unit=MM):
        """ Calculate path lengths per tool.

        This function only sums actual cut lengths, and ignores travel lengths that the tool is doing without cutting to
        get from one object to another. Travel lengths depend on the CAM program's path planning, which highly depends
        on panelization and other factors. Additionally, an EDA tool will not even attempt to minimize travel distance
        as that's not its job.

        :param unit: :py:class:`.LengthUnit` or str (``'mm'`` or ``'inch'``). Unit to use for return value. Default: mm

        :returns: ``{ tool: float(path length) }``
        :rtype dict:
        """
        lengths = {}
        tool = None
        for obj in sorted(self.objects, key=lambda obj: obj.tool):
            if tool != obj.tool:
                tool = obj.tool
                lengths[tool] = 0

            lengths[tool] += obj.curve_length(unit)
        return lengths

    def hit_count(self):
        """ Calculate the number of objects per tool.

        :rtype: collections.Counter
        """
        return Counter(obj.tool for obj in self.objects)

    def drill_sizes(self, unit=MM):
        """ Return a sorted list of all tool diameters found in this file.

        :param unit: :py:class:`.LengthUnit` or str (``'mm'`` or ``'inch'``). Unit to use for return values. Default: mm

        :returns: list of floats, sorted smallest to largest diameter.
        :rtype: list
        """
        # use equivalent_width for unit conversion
        return sorted({ obj.tool.equivalent_width(unit) for obj in self.objects })

    def drills(self):
        """ Return all drilled hole objects in this file.

        :returns: list of :py:class:`.Flash` instances
        :rtype: list
        """
        return (obj for obj in self.objects if isinstance(obj, Flash))

    def slots(self):
        """ Return all milled slot objects in this file.

        :returns: list of :py:class:`~.graphic_objects.Line` or :py:class:`~.graphic_objects.Arc` instances
        :rtype: list
        """
        return (obj for obj in self.objects if not isinstance(obj, Flash))


class ProgramState(Enum):
    """ Internal helper class used to track Excellon program state (i.e. G05/G06 command state). """
    HEADER = 0
    DRILLING = 1
    ROUTING = 2
    FINISHED = 3


class ExcellonParser(object):
    """ Internal helper class that contains all the actual Excellon format parsing logic. """

    def __init__(self, settings=None, external_tools=None):
        # NOTE XNC files do not contain an explicit number format specification, but all values have decimal points.
        # Thus, we set the default number format to (None, None). If the file does not contain an explicit specification
        # and FileSettings.parse_gerber_value encounters a number without an explicit decimal point, it will throw a
        # SyntaxError. In case of e.g. Allegro files where the number format and other options are specified separately
        # from the excellon file, the caller must pass in an already filled-out FileSettings object.
        if settings is None:
            self.settings = FileSettings(number_format=(None, None), zeros='leading')
        else:
            self.settings = settings
        self.program_state = None
        self.interpolation_mode = InterpMode.LINEAR
        self.tools = {}
        self.objects = []
        self.active_tool = None
        self.pos = 0, 0
        self.drill_down = False
        self.is_plated = None
        self.comments = []
        self.generator_hints = []
        self.lineno = None
        self.filename = None
        self.external_tools = external_tools or {}
        self.found_kicad_format_comment = False

    def warn(self, msg):
        warnings.warn(f'{self.filename}:{self.lineno} "{self.line}": {msg}', SyntaxWarning)

    def do_parse(self, data, filename=None):
        # filename arg is for error messages
        self.filename = filename = filename or '<unknown>'

        leftover = None
        for lineno, line in enumerate(data.splitlines(), start=1):
            line = line.strip()
            self.lineno, self.line = lineno, line # for warnings

            if not line:
                continue

            # Coordinates of G00 and G01 may be on the next line
            if line == 'G00' or line == 'G01':
                if leftover:
                    self.warn('Two consecutive G00/G01 commands without coordinates. Ignoring first.')
                leftover = line
                continue

            if leftover:
                line = leftover + line
                leftover = None

            if line and self.program_state == ProgramState.FINISHED:
                self.warn('Commands found following end of program statement.')
            # TODO check first command in file is "start of header" command.

            try:
                if not self.exprs.handle(self, line):
                    raise ValueError('Unknown excellon statement:', line)
            except Exception as e:
                raise SyntaxError(f'{filename}:{lineno} "{line}": {e}') from e

    exprs = RegexMatcher()

    # NOTE: These must be kept before the generic comment handler at the end of this class so they match first.
    @exprs.match(r';T(?P<index1>[0-9]+) Holesize (?P<index2>[0-9]+)\. = (?P<diameter>[0-9/.]+) Tolerance = \+[0-9/.]+/-[0-9/.]+ (?P<plated>PLATED|NON_PLATED|OPTIONAL) (?P<unit>MILS|MM) Quantity = [0-9]+')
    def parse_allegro_tooldef(self, match):
        # NOTE: We ignore the given tolerances here since they are non-standard.
        self.program_state = ProgramState.HEADER # TODO is this needed? we need a test file.
        self.generator_hints.append('allegro')

        if (index := int(match['index1'])) != int(match['index2']): # index1 has leading zeros, index2 not.
            raise SyntaxError('BUG: Allegro excellon tool def has mismatching tool indices. Please file a bug report on our issue tracker and provide this file!')

        if index in self.tools:
            self.warn('Re-definition of tool index {index}, overwriting old definition.') 

        # NOTE: We map "optionally" plated holes to plated holes for API simplicity. If you hit a case where that's a
        # problem, please raise an issue on our issue tracker, explain why you need this and provide an example file.
        is_plated = None if match['plated'] is None else (match['plated'] in ('PLATED', 'OPTIONAL'))

        diameter = float(match['diameter'])

        if match['unit'] == 'MILS':
            diameter /= 1000
            unit = Inch
        else:
            unit = MM

        if unit != self.settings.unit:
            self.warn('Allegro Excellon drill file tool definitions in {unit.name}, but file parameters say the '
                    'file should be in {settings.unit.name}. Please double-check that this is correct, and if it is, '
                    'please raise an issue on our issue tracker.')

        self.tools[index] = ExcellonTool(diameter=diameter, plated=is_plated, unit=unit)

    # Searching Github I found that EasyEDA has two different variants of the unit specification here.
    @exprs.match(';Holesize (?P<index>[0-9]+) = (?P<diameter>[.0-9]+) (?P<unit>INCH|inch|METRIC|mm)')
    def parse_easyeda_tooldef(self, match):
        unit = Inch if match['unit'].lower() == 'inch' else MM
        tool = ExcellonTool(diameter=float(match['diameter']), unit=unit, plated=self.is_plated)

        if (index := int(match['index'])) in self.tools:
            self.warn('Re-definition of tool index {index}, overwriting old definition.') 

        self.tools[index] = tool
        self.generator_hints.append('easyeda')

    @exprs.match('T([0-9]+)(([A-Z][.0-9]+)+)') # Tool definition: T** with at least one parameter
    def parse_normal_tooldef(self, match):
        # We ignore parameters like feed rate or spindle speed that are not used for EDA -> CAM file transfer. This is
        # not a parser for the type of Excellon files a CAM program sends to the machine.

        if (index := int(match[1])) in self.tools:
            self.warn('Re-definition of tool index {index}, overwriting old definition.') 

        params = { m[0]: self.settings.parse_gerber_value(m[1:]) for m in re.findall('[BCFHSTZ][.0-9]+', match[2]) }

        self.tools[index] = ExcellonTool(diameter=params.get('C'), plated=self.is_plated,
                unit=self.settings.unit)

        if set(params.keys()) == set('TFSC'):
            self.generator_hints.append('target3001') # target files look like altium files without the comments

        if len(self.tools) >= 3 and list(self.tools.keys()) == reversed(sorted(self.tools.keys())):
            self.generator_hints.append('geda')

    @exprs.match('T([0-9]+)')
    def parse_tool_selection(self, match):
        index = int(match[1])

        if index == 0: # T0 is used as END marker, just ignore
            return
        elif index not in self.tools:
            if not self.tools and index in self.external_tools:
                # allegro is just wonderful.
                self.warn(f'Undefined tool index {index} selected. We found an allegro drill log file next to this, so '
                            'we will use tool definitions from there.')
                self.active_tool = self.external_tools[index]

            else:
                raise SyntaxError(f'Undefined tool index {index} selected.')

        else:
            self.active_tool = self.tools[index]

    coord = lambda name: fr'(?:{name}\+?(-?)([0-9]*\.?[0-9]*))?'
    xy_coord = coord('X') + coord('Y')
    xyaij_coord = xy_coord + coord('A') + coord('I') + coord('J')

    @exprs.match(r'R([0-9]+)' + xy_coord)
    def handle_repeat_hole(self, match):
        if self.program_state == ProgramState.HEADER:
            return

        count, x_s, x, y_s, y = match.groups()
        dx = self.settings.parse_gerber_value(x) or 0
        if x_s:
            dx = -dx
        dy = self.settings.parse_gerber_value(y) or 0
        if y_s:
            dy = -dy

        for i in range(int(count)):
            self.pos = (self.pos[0] + dx, self.pos[1] + dy)
            # FIXME fix API below
            if not self.ensure_active_tool():
                return

            self.objects.append(Flash(*self.pos, self.active_tool, unit=self.settings.unit))

    def header_command(name):
        def wrap(fun):
            @functools.wraps(fun)
            def wrapper(self, *args, **kwargs):
                nonlocal name
                if self.program_state is None:
                    self.warn(f'{name} header statement found before start of header')
                elif self.program_state != ProgramState.HEADER:
                    self.warn(f'{name} header statement found after end of header')
                fun(self, *args, **kwargs)
            return wrapper
        return wrap

    @exprs.match('M48')
    def handle_begin_header(self, match):
        if self.program_state == ProgramState.HEADER:
            # It seems that only fritzing puts both a '%' start of header thingy and an M48 statement at the beginning
            # of the file.
            self.generator_hints.append('fritzing')
        elif self.program_state is not None:
            self.warn(f'M48 "header start" statement found in the middle of the file, currently in {self.program_state}')
        self.program_state = ProgramState.HEADER

    @exprs.match('M95')
    @header_command('M95')
    def handle_end_header(self, match):
        self.program_state = ProgramState.DRILLING

    @exprs.match('M15')
    def handle_drill_down(self, match):
        self.drill_down = True

    @exprs.match('M16|M17')
    def handle_drill_up(self, match):
        self.drill_down = False


    @exprs.match('M30|M00')
    def handle_end_of_program(self, match):
        if self.program_state in (None, ProgramState.HEADER):
            self.warn('M30 statement found before end of header.')
        self.program_state = ProgramState.FINISHED
        # TODO: maybe add warning if this is followed by other commands.

        if match[0] == 'M00':
            self.generator_hints.append('zuken')

    def do_move(self, coord_groups):
        x_s, x, y_s, y = coord_groups

        if self.settings.number_format == (None, None) and '.' not in x:
            # TARGET3001! exports zeros as "00" even when it uses an explicit decimal point everywhere else.
            if x != '00':
                raise SyntaxError('No number format set and value does not contain a decimal point. If this is an Allegro '
                    'Excellon drill file make sure either nc_param.txt or ncdrill.log ends up in the same folder as '
                    'it, because Allegro does not include this critical information in their Excellon output. If you '
                    'call this through ExcellonFile.from_string, you must manually supply from_string with a '
                    'FileSettings object from excellon.parse_allegro_ncparam.')

        x = self.settings.parse_gerber_value(x)
        if x_s:
            x = -x
        y = self.settings.parse_gerber_value(y)
        if y_s:
            y = -y

        old_pos = self.pos

        if self.settings.is_absolute:
            if x is not None:
                self.pos = (x, self.pos[1])
            if y is not None:
                self.pos = (self.pos[0], y)
        else: # incremental
            if x is not None:
                self.pos = (self.pos[0]+x, self.pos[1])
            if y is not None:
                self.pos = (self.pos[0], self.pos[1]+y)

        return old_pos, self.pos

    @exprs.match('G00' + xy_coord)
    def handle_start_routing(self, match):
        if self.program_state is None:
            self.warn('Routing mode command found before header.')
        self.program_state = ProgramState.ROUTING
        self.do_move(match.groups())

    @exprs.match('%')
    def handle_rewind_shorthand(self, match):
        if self.program_state is None:
            self.program_state = ProgramState.HEADER
        elif self.program_state is ProgramState.HEADER:
            self.program_state = ProgramState.DRILLING
        # FIXME handle rewind start

    @exprs.match('G05')
    def handle_drill_mode(self, match):
        self.drill_down = False
        self.program_state = ProgramState.DRILLING

    def ensure_active_tool(self):
        if self.active_tool:
            return self.active_tool
        
        self.warn('Routing command found before first tool definition.')
        return None

    @exprs.match('(G01|G02|G03)' + xyaij_coord)
    def handle_linear_mode(self, match):
        mode, *coord_groups = match.groups()
        if mode == 'G01':
            self.interpolation_mode = InterpMode.LINEAR
        else:
            clockwise = (mode == 'G02')
            self.interpolation_mode = InterpMode.CIRCULAR_CW if clockwise else InterpMode.CIRCULAR_CCW

        self.do_interpolation(coord_groups)
    
    def do_interpolation(self, coord_groups):
        x_s, x, y_s, y, a_s, a, i_s, i, j_s, j = coord_groups

        start, end = self.do_move((x_s, x, y_s, y))

        if self.program_state != ProgramState.ROUTING:
            return

        if not self.drill_down or not (x or y) or not self.ensure_active_tool():
            return

        if self.interpolation_mode == InterpMode.LINEAR:
            if a or i or j:
                self.warn('A/I/J arc coordinates found in linear mode.')

            self.objects.append(Line(*start, *end, self.active_tool, unit=self.settings.unit))

        else:
            if (x or y) and not (a or i or j):
                self.warn('Arc without radius found.')

            clockwise = (self.interpolation_mode == InterpMode.CIRCULAR_CW)
            
            if a: # radius given
                if i or j:
                    self.warn('Arc without both radius and center specified.')

                # Convert endpoint-radius-endpoint notation to endpoint-center-endpoint notation. We always use the
                # smaller arc here.
                # from https://math.stackexchange.com/a/1781546
                if a_s:
                    raise ValueError('Negative arc radius given')
                r = settings.parse_gerber_value(a)
                x1, y1 = start
                x2, y2 = end
                dx, dy = (x2-x1)/2, (y2-y1)/2
                x0, y0 = x1+dx, y1+dy
                f = math.hypot(dx, dy) / math.sqrt(r**2 - a**2)
                if clockwise:
                    cx = x0 + f*dy
                    cy = y0 - f*dx
                else:
                    cx = x0 - f*dy
                    cy = y0 + f*dx
                i, j = cx-start[0], cy-start[1]

            else: # explicit center given
                i = settings.parse_gerber_value(i)
                if i_s:
                    i = -i
                j = settings.parse_gerber_value(j)
                if j_s:
                    j = -i

            self.objects.append(Arc(*start, *end, i, j, True, self.active_tool, unit=self.settings.unit))

    @exprs.match(r'(M71|METRIC|M72|INCH)(,LZ|,TZ)?(,0*\.0*)?')
    def parse_easyeda_format(self, match):
        metric = match[1] in ('METRIC', 'M71')

        self.settings.unit = MM if metric else Inch

        if match[2]:
            self.settings.zeros = 'trailing' if match[2] == ',LZ' else 'leading'

        # Newer EasyEDA exports have this in an altium-like FILE_FORMAT comment instead. Some files even have both.
        # This is used by newer autodesk eagles, fritzing and diptrace
        if match[3]:
            integer, _, fractional = match[3][1:].partition('.')
            self.settings.number_format = len(integer), len(fractional)

        elif self.settings.number_format == (None, None) and not metric and not self.found_kicad_format_comment:
            self.warn('Using implicit number format from bare "INCH" statement. This is normal for Fritzing, Diptrace, Geda and pcb-rnd.')
            self.settings.number_format = (2,4)
    
    @exprs.match('G90')
    @header_command('G90')
    def handle_absolute_mode(self, match):
        self.settings.notation = 'absolute'

    @exprs.match('G93' + xy_coord)
    def handle_absolute_mode(self, match):
        _x_s, x, _y_s, y = match.groups()
        if int(x or 0) != 0 or int(y or 0) != 0:
            # Siemens tooling likes to include a meaningless G93X0Y0 after its header.
            raise NotImplementedError('G93 zero set command is not supported.')
        self.generator_hints.append('siemens')

    @exprs.match('ICI,?(ON|OFF)')
    def handle_incremental_mode(self, match):
        self.settings.notation = 'absolute' if match[1] == 'OFF' else 'incremental'

    @exprs.match('(FMAT|VER),?([0-9]*)')
    def handle_command_format(self, match):
        if match[1] == 'FMAT':
            # We do not support integer/fractional decimals specification via FMAT because that's stupid. If you need this,
            # please raise an issue on our issue tracker, provide a sample file and tell us where on earth you found that
            # file.
            if match[2] not in ('', '2'):
                raise SyntaxError(f'Unsupported FMAT format version {match[2]}')

        else: # VER
            self.generator_hints.append('zuken')

    @exprs.match(r'G40|G41|G42|F[0-9]+')
    def handle_unhandled(self, match):
        self.warn(f'{match[0]} excellon command intended for CAM tools found in EDA file.')

    @exprs.match(xyaij_coord)
    def handle_bare_coordinate(self, match):
        # Yes, drills in the header doesn't follow the specification, but it there are many files like this.
        if self.program_state in (ProgramState.DRILLING, ProgramState.HEADER):
            _start, end = self.do_move(match.groups()[:4])

            if not self.ensure_active_tool():
                return

            self.objects.append(Flash(*end, self.active_tool, unit=self.settings.unit))

        elif self.program_state == ProgramState.ROUTING:
            # Bare coordinates for routing also seem illegal, but Siemens actually uses these.
            # Example file: siemens/80101_0125_F200_ContourPlated.ncd
            self.do_interpolation(match.groups())

        else:
            self.warn('Bare coordinate after end of file')

    @exprs.match(r'DETECT,ON|ATC,ON|M06')
    def parse_zuken_legacy_statements(self, match):
        self.generator_hints.append('zuken')

    @exprs.match(r'; Format\s*: ([0-9]+\.[0-9]+) / (Absolute|Incremental) / (Inch|MM) / (Leading|Trailing)')
    def parse_siemens_format(self, match):
        x, _, y = match[1].partition('.')
        self.settings.number_format = int(x), int(y)
        # NOTE: Siemens files seem to always contain both this comment and an explicit METRIC/INC statement. However,
        # the meaning of "leading" and "trailing" is swapped in both: When this comment says leading, we get something
        # like "INCH,TZ".
        self.settings.notation = match[2].lower()
        self.settings.unit = to_unit(match[3])
        self.settings.zeros = {'Leading': 'trailing', 'Trailing': 'leading'}[match[4]]
        self.generator_hints.append('siemens')

    @exprs.match('; Contents: (Thru|.*) / (Drill|Mill) / (Plated|Non-Plated)')
    def parse_siemens_meta(self, match):
        self.is_plated = (match[3] == 'Plated')
        self.generator_hints.append('siemens')

    @exprs.match(';FILE_FORMAT=([0-9]:[0-9])')
    def parse_altium_easyeda_number_format_comment(self, match):
        # Altium or newer EasyEDA exports
        x, _, y = match[1].partition(':')
        self.settings.number_format = int(x), int(y)

    @exprs.match(';Layer: (.*)')
    def parse_easyeda_layer_name(self, match):
        # EasyEDA embeds the layer name in a comment. EasyEDA uses separate files for plated/non-plated. The (default?)
        # layer names are: "Drill PTH", "Drill NPTH"
        self.is_plated = 'NPTH' not in match[1]
        self.generator_hints.append('easyeda')

    @exprs.match(';TYPE=(PLATED|NON_PLATED)')
    def parse_altium_composite_plating_comment(self, match):
        # These can happen both before a tool definition and before a tool selection statement.
        # FIXME make sure we do the right thing in both cases.
        self.is_plated = (match[1] == 'PLATED')

    @exprs.match(';(Layer_Color=[-+0-9a-fA-F]*)')
    def parse_altium_layer_color(self, match):
        self.generator_hints.append('altium')
        self.comments.append(match[1])
    
    @exprs.match(';HEADER:')
    def parse_allegro_start_of_header(self, match):
        self.program_state = ProgramState.HEADER
        self.generator_hints.append('allegro')

    @exprs.match(r';GenerationSoftware,Autodesk,EAGLE,.*\*%')
    def parse_eagle_version_header(self, match):
        # NOTE: Only newer eagles export drills as XNC files. Older eagles produce an aperture-only gerber file called
        # "profile.gbr" instead.
        self.generator_hints.append('eagle')

    @exprs.match(';EasyEDA .*')
    def parse_easyeda_version_header(self, match):
        self.generator_hints.append('easyeda')

    @exprs.match(';DRILL .*KiCad .*')
    def parse_kicad_version_header(self, match):
        self.generator_hints.append('kicad')
    
    @exprs.match(';FORMAT={([-0-9]+:[-0-9]+) ?/ (.*) / (inch|.*) / decimal}')
    def parse_kicad_number_format_comment(self, match):
        x, _, y = match[1].partition(':')
        x = None if x == '-' else int(x)
        y = None if y == '-' else int(y)
        self.settings.number_format = x, y
        self.found_kicad_format_comment = True
        self.settings.notation = match[2]
        self.settings.unit = Inch if match[3] == 'inch' else MM

    @exprs.match(';(.*)')
    def parse_comment(self, match):
        self.comments.append(match[1].strip())

        if all(cmt.startswith(marker)
                for cmt, marker in zip(reversed(self.comments), ['Version', 'Job', 'User', 'Date'])):
            self.generator_hints.append('siemens')




#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2023 Jan Sebastian Götte <gerbonara@jaseg.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import math
import click
import dataclasses
import re
import warnings
import json
import itertools
from pathlib import Path

from .utils import MM, Inch
from .cam import FileSettings
from .rs274x import GerberFile
from . import layers as lyr
from . import __version__
from .cad.kicad import schematic as kc_schematic
from .cad.kicad import tmtheme


def _print_version(ctx, param, value):
    if value and not ctx.resilient_parsing:
        click.echo(f'Version {__version__}')
        ctx.exit()


def _apply_transform(transform, unit, layer_or_stack):
    def translate(x, y):
        layer_or_stack.offset(x, y, unit)

    def scale(factor):
        """ Scale layer by a given factor, e.g. 1.0 for no change, 2.0 to double all coordinates in both axes. Note that
        we only offer uniform scaling with a single factor applied along both coordinate axes because anything else
        would not be possible with arbitrary Gerber apertures, and definitely mess up holes. We could still do this, but
        the result would almost certainly not be what the user is looking for.

        The main reason why this function might make sense is to fix up boards exported as G-code by programs that
        aren't EDA tools and that for whatever reason ended up exporting in a weird unit."""
        layer_or_stack.scale(factor)

    def rotate(angle, cx=0, cy=0):
        layer_or_stack.rotate(math.radians(angle), cx, cy, unit)

    (x_min, y_min), (x_max, y_max) = layer_or_stack.bounding_box(unit, default=((0, 0), (0, 0)))
    width, height = x_max - x_min, y_max - y_min

    def origin():
        translate(-x_min, -y_min)

    def center():
        translate(-x_min-width/2, -y_min-height/2)

    exec(transform, {key: value for key, value in math.__dict__.items() if not key.startswith('_')}, locals())


class Coordinate(click.ParamType):
    name = 'coordinate'

    def __init__(self, dimension=2):
        self.dimension = dimension
    
    def convert(self, value, param, ctx):
        try:
            coords = [float(e) for e in value.split(',')]
            if len(coords) != self.dimension:
                raise ValueError()
            return coords

        except ValueError:
            self.fail(f'{value!r} is not a valid coordinate. A coordinate consists of exactly {self.dimension} comma-separate floating-point numbers.')

class Rotation(click.ParamType):
    name = 'rotation'

    def convert(self, value, param, ctx):
        try:
            coords = [float(e) for e in value.split(',')]
            if len(coords) not in (1, 3):
                raise ValueError()

            theta, x, y, *_rest = *coords, 0, 0
            return theta, x, y

        except ValueError:
            self.fail(f'{value!r} is not a valid rotation. A rotation is either a floating point angle ("[theta]"), or the same angle followed by comma-separated X and Y coordinates of the rotation center ("[theta],[cx],[cy]").')


class Unit(click.Choice):
    name = 'unit'

    def __init__(self):
        super().__init__(['metric', 'us-customary'])

    def convert(self, value, param, ctx):
        value = super().convert(value, param, ctx)
        return MM if value == 'metric' else Inch


class NamingScheme(click.Choice):
    name = 'naming_scheme'

    def __init__(self):
        super().__init__([n for n in dir(lyr.NamingScheme) if not n.startswith('_')])

    def convert(self, value, param, ctx):
        return getattr(lyr.NamingScheme, super().convert(value, param, ctx))


@click.group()
@click.option('--version', is_flag=True, callback=_print_version, expose_value=False, is_eager=True)
def cli():
    """ The gerbonara CLI allows you to analyze, render, modify and merge both individual Gerber or Excellon files as
    well as sets of those files """
    pass


@cli.group('kicad')
def kicad_group():
    pass


@kicad_group.group('schematic')
def schematic_group():
    pass


@schematic_group.command()
@click.argument('inpath', type=click.Path(exists=True))
@click.argument('theme', type=click.Path(exists=True))
@click.argument('outfile', type=click.File('w'), default='-')
def render(inpath, theme, outfile):
    sch = kc_schematic.Schematic.open(inpath)
    cs = tmtheme.TmThemeSchematic(Path(theme).read_text())
    with outfile as f:
        f.write(str(sch.to_svg(cs)))


@cli.command()
@click.option('--warnings', 'format_warnings', type=click.Choice(['default', 'ignore', 'once']), default='default',
              help='''Enable or disable file format warnings during parsing (default: on)''')
@click.option('--version', is_flag=True, callback=_print_version, expose_value=False, is_eager=True)
@click.option('-m', '--input-map', type=click.Path(exists=True, path_type=Path), help='''Extend or override layer name
              mapping with name map from JSON file. The JSON file must contain a single JSON dict with an arbitrary
              number of string: string entries. The keys are interpreted as regexes applied to the filenames via
              re.fullmatch, and each value must either be the string "ignore" to remove this layer from previous
              automatic guesses, or a gerbonara layer name such as "top copper", "inner_2 copper" or "bottom silk".''')
@click.option('--use-builtin-name-rules/--no-builtin-name-rules', default=True, help='''Disable built-in layer name
              rules and use only rules given by --input-map''')
@click.option('--force-zip', is_flag=True, help='''Force treating input path as a zip file (default: guess file type
              from extension and contents)''')
@click.option('--top', 'side', flag_value='top', help='Render top side')
@click.option('--bottom', 'side', flag_value='bottom', help='Render top side')
@click.option('--command-line-units', type=Unit(), help='''Units for values given in other options. Default:
              millimeter''')
@click.option('--margin', type=float, default=0.0, help='Add space around the board inside the viewport')
@click.option('--force-bounds', help='Force SVG bounding box to value given as "min_x,min_y,max_x,max_y"')
@click.option('--inkscape/--standard-svg', default=True, help='Export in Inkscape SVG format with layers and stuff.')
@click.option('--pretty/--no-filters', default=True, help='''Export pseudo-realistic render using filters (default) or
              just stack up layers using given colorscheme. In "--no-filters" mode, by default all layers are exported
              unless either "--top" or "--bottom" is given.''')
@click.option('--drills/--no-drills', default=True, help='''Include (default) or exclude drills ("--no-filters" only!)''')
@click.option('--colorscheme', type=click.Path(exists=True, path_type=Path), help='''Load colorscheme from given JSON
              file. The JSON file must contain a single dict with keys copper, silk, mask, paste, drill and outline.
              Each key must map to a string containing either a normal 6-digit hex color with leading hash sign, or an
              8-digit hex color with leading hash sign, where the last two digits set the layer's alpha value (opacity),
              with FF being completely opaque, and 00 being invisibly transparent.''')
@click.argument('inpath', type=click.Path(exists=True))
@click.argument('outfile', type=click.File('w'), default='-')
def render(inpath, outfile, format_warnings, input_map, use_builtin_name_rules, force_zip, side, drills,
           command_line_units, margin, force_bounds, inkscape, pretty, colorscheme):
    """ Render a gerber file, or a directory or zip of gerber files into an SVG file. """

    overrides = json.loads(input_map.read_bytes()) if input_map else None
    with warnings.catch_warnings():
        warnings.simplefilter(format_warnings)
        if force_zip:
            stack = lyr.LayerStack.open_zip(inpath, overrides=overrides, autoguess=use_builtin_name_rules)
        else:
            stack = lyr.LayerStack.open(inpath, overrides=overrides, autoguess=use_builtin_name_rules)

    if force_bounds:
        min_x, min_y, max_x, max_y = list(map(float, force_bounds.split(',')))
        force_bounds = (min_x, min_y), (max_x, max_y)

    if colorscheme:
        colorscheme = json.loads(colorscheme.read_text())

    if pretty:
        svg = stack.to_pretty_svg(side='bottom' if side == 'bottom' else 'top', margin=margin,
                                              arg_unit=(command_line_units or MM),
                          svg_unit=MM, force_bounds=force_bounds, inkscape=inkscape, colors=colorscheme)
    else:
        svg = stack.to_svg(side_re=side or '.*', margin=margin, drills=drills, arg_unit=(command_line_units or MM),
                          svg_unit=MM, force_bounds=force_bounds, colors=colorscheme)
    outfile.write(str(svg))


@cli.command()
@click.option('--version', is_flag=True, callback=_print_version, expose_value=False, is_eager=True)
@click.option('--warnings', 'format_warnings', type=click.Choice(['default', 'ignore', 'once']), default='default',
              help='''Enable or disable file format warnings during parsing (default: on)''')
@click.option('-t', '--transform', help='''Execute python transformation script on input. You have access to the
              functions translate(x, y), scale(factor) and rotate(angle, center_x?, center_y?), the bounding box
              variables x_min, y_min, x_max, y_max, width and height, and everything from python\'s built-in math module
              (e.g. pi, sqrt, sin). As convenience methods, center() and origin() are provided to center the board resp.
              move its bottom-left corner to the origin. Coordinates are given in --command-line-units, angles in
              degrees, and scale as a scale factor (as opposed to a percentage). Example: "translate(-10, 0); rotate(45,
              0, 5)"''')
@click.option('--command-line-units', type=Unit(), help='''Units for values given in other options. Default:
              millimeter''')
@click.option('-n', '--number-format', help='''Override number format to use during export in "[integer digits].[decimal
              digits]" notation, e.g. "2.6".''')
@click.option('-u', '--units', type=Unit(), help='Override export file units')
@click.option('-z', '--zero-suppression', type=click.Choice(['off', 'leading', 'trailing']), help='''Override export
              zero suppression setting. Note: The meaning of this value is like in the Gerber spec for both Gerber and
              Excellon files!''')
@click.option('--keep-comments/--drop-comments', help='''Keep gerber comments. Note: Comments will be prepended to the
              start of file, and will not occur in their old position.''')
@click.option('--reuse-input-settings', 'output_format', flag_value='reuse', help='''Use the same export settings as the
              input file instead of sensible defaults.''')
@click.option('--default-settings', 'output_format', default=True, flag_value='defaults', help='''Use sensible defaults
              for the output file format settings (default).''')
@click.option('--input-number-format', help='Override number format of input file (mostly useful for Excellon files)')
@click.option('--input-units', type=Unit(), help='Override units of input file')
@click.option('--input-zero-suppression', type=click.Choice(['off', 'leading', 'trailing']), help='''Override zero
              suppression setting of input file''')
@click.argument('infile')
@click.argument('outfile')
def rewrite(transform, command_line_units, number_format, units, zero_suppression, keep_comments, output_format,
            input_number_format, input_units, input_zero_suppression, infile, outfile, format_warnings):
    """ Parse a single gerber file, apply transformations, and re-serialize it into a new gerber file. Without
    transformations, this command can be used to convert a gerber file to use different settings (e.g. units,
    precision), but can also be used to "normalize" gerber files in a weird format into a more standards-compatible one
    as gerbonara's gerber parser is significantly more robust for weird inputs than others. """

    input_settings = FileSettings()
    if input_number_format:
        a, _, b = input_number_format.partition('.')
        input_settings.number_format = (int(a), int(b))

    if input_zero_suppression:
        input_settings.zeros = None if input_zero_suppression == 'off' else input_zero_suppression

    input_settings.unit = input_units

    with warnings.catch_warnings():
        warnings.simplefilter(format_warnings)
        f = GerberFile.open(infile, override_settings=input_settings)

    if transform:
        _apply_transform(transform, command_line_units or MM, f)

    output_format = f.import_settings if output_format == 'reuse' else FileSettings.defaults()
    if number_format:
        a, _, b = number_format.partition('.')
        output_format.number_format = (int(a), int(b))

    if units:
        output_format.unit = units

    if zero_suppression:
        output_format.zeros = None if zero_suppression == 'off' else zero_suppression

    f.save(outfile, output_format, not keep_comments)


@cli.command()
@click.option('--version', is_flag=True, callback=_print_version, expose_value=False, is_eager=True)
@click.option('-m', '--input-map', type=click.Path(exists=True, path_type=Path), help='''Extend or override layer name
              mapping with name map from JSON file. The JSON file must contain a single JSON dict with an arbitrary
              number of string: string entries. The keys are interpreted as regexes applied to the filenames via
              re.fullmatch, and each value must either be the string "ignore" to remove this layer from previous
              automatic guesses, or a gerbonara layer name such as "top copper", "inner_2 copper" or "bottom silk".''')
@click.option('--use-builtin-name-rules/--no-builtin-name-rules', default=True, help='''Disable built-in layer name
              rules and use only rules given by --input-map''')
@click.option('--warnings', 'format_warnings', type=click.Choice(['default', 'ignore', 'once']), default='default',
              help='''Enable or disable file format warnings during parsing (default: on)''')
@click.option('--units', type=Unit(), help='Units for values given in other options. Default: millimeter')
@click.option('-n', '--number-format', help='''Override number format to use during export in
              "[integer digits].[decimal digits]" notation, e.g. "2.6".''')
@click.option('--reuse-input-settings', 'output_format', flag_value='reuse', help='''Use the same export settings as the
              input file instead of sensible defaults.''')
@click.option('--default-settings', 'output_format', default=True, flag_value='defaults', help='''Use sensible defaults
              for the output file format settings (default).''')
@click.option('--force-zip', is_flag=True, help='''Force treating input path as a zip file (default: guess file type
              from extension and contents)''')
@click.option('--output-naming-scheme', type=NamingScheme(), help=f'''Name output files according to the selected naming
              scheme instead of keeping the old file names.''')
@click.argument('transform')
@click.argument('inpath')
@click.argument('outpath')
def transform(transform, units, output_format, inpath, outpath,
            format_warnings, input_map, use_builtin_name_rules, output_naming_scheme):
    """ Transform all gerber files in a given directory or zip file using the given python transformation script.
        
        In the python transformation script you have access to the functions translate(x, y), scale(factor) and
        rotate(angle, center_x?, center_y?), the bounding box variables x_min, y_min, x_max, y_max, width and height,
        and everything from python\'s built-in math module (e.g. pi, sqrt, sin). As convenience methods, center() and
        origin() are provided to center the board resp. move its bottom-left corner to the origin. Coordinates are given
        in --command-line-units, angles in degrees, and scale as a scale factor (as opposed to a percentage). Example:
        "translate(-10, 0); rotate(45, 0, 5)"''')
    """

    overrides = json.loads(input_map.read_bytes()) if input_map else None
    with warnings.catch_warnings():
        warnings.simplefilter(format_warnings)
        if force_zip:
            stack = lyr.LayerStack.open_zip(path, overrides=overrides, autoguess=use_builtin_name_rules)
        else:
            stack = lyr.LayerStack.open(path, overrides=overrides, autoguess=use_builtin_name_rules)

    _apply_transform(transform, units, stack)

    output_format = None if output_format == 'reuse' else FileSettings.defaults()
    stack.save_to_directory(outpath, naming_scheme=output_naming_scheme or {},
                            gerber_settings=output_format,
                            excellon_settings=dataclasses.replace(output_format, zeros=None))


@cli.command()
@click.option('--version', is_flag=True, callback=_print_version, expose_value=False, is_eager=True)
@click.option('--command-line-units', type=Unit(), help='''Units for values given in --transform. Default:
              millimeter''')
@click.option('--warnings', 'format_warnings', type=click.Choice(['default', 'ignore', 'once']), default='default',
              help='''Enable or disable file format warnings during parsing (default: on)''')
@click.option('--offset', multiple=True, type=Coordinate(), help="""Offset for the n'th file as a "x,y" string in unit
              given by --command-line-units (default: millimeter). Can be given multiple times, and the first option
              affects the first input, the second option affects the second input, and so on.""")
@click.option('--rotation', multiple=True, type=Rotation(), help="""Rotation for the n'th file in degrees clockwise,
              optionally followed by comma-separated rotation center X and Y coordinates. Can be given multiple times,
              and the first option affects the first input, the second option affects the second input, and so on.""")
@click.option('-m', '--input-map', type=click.Path(exists=True, path_type=Path), multiple=True, help='''Extend or
              override layer name mapping with name map from JSON file. This option can be given multiple times, in
              which case the n'th option affects only the n'th input, like with --offset and --rotation. The JSON file
              must contain a single JSON dict with an arbitrary number of string: string entries. The keys are
              interpreted as regexes applied to the filenames via re.fullmatch, and each value must either be the string
              "ignore" to remove this layer from previous automatic guesses, or a gerbonara layer name such as "top
              copper", "inner_2 copper" or "bottom silk".''')
@click.option('--reuse-input-settings', 'output_format', flag_value='reuse', help='''Use the same export settings as the
              input file instead of sensible defaults.''')
@click.option('--default-settings', 'output_format', default=True, flag_value='defaults', help='''Use sensible defaults
              for the output file format settings (default).''')
@click.option('--output-naming-scheme', type=NamingScheme(), help=f'''Name output files according to the selected naming
              scheme instead of keeping the old file names of the first input.''')
@click.option('--output-board-name', help=f'''Override board name used with --output-naming-scheme''')
@click.option('--use-builtin-name-rules/--no-builtin-name-rules', default=True, help='''Disable built-in layer name
              rules and use only rules given by --input-map''')
@click.argument('inpath', nargs=-1, type=click.Path(exists=True, path_type=Path))
@click.argument('outpath', type=click.Path(path_type=Path))
def merge(inpath, outpath, offset, rotation, input_map, command_line_units, output_format, output_naming_scheme,
          output_board_name, format_warnings, use_builtin_name_rules):
    """ Merge multiple single Gerber or Excellon files, or multiple stacks of Gerber files, into one. Hint: When used
    with only one input, this command "normalizes" the input, converting all files to a well-defined, widely supported
    Gerber subset with sane settings. When a --output-naming-scheme is given, it additionally renames all files to a
    standardized naming convention. """
    if not inpath:
        return

    target = None
    for p, offset, rotation, input_map in itertools.zip_longest(inpath, offset, rotation, input_map):
        if p is None:
            raise click.UsageError('More --offset, --rotation or --input-map options than input files')

        offset = offset or (0, 0)
        theta, cx, cy = rotation or (0, 0, 0)

        overrides = json.loads(input_map.read_bytes()) if input_map else None
        with warnings.catch_warnings():
            warnings.simplefilter(format_warnings)

            stack = lyr.LayerStack.open(p, overrides=overrides, autoguess=use_builtin_name_rules)

            if not math.isclose(offset[0], 0, abs_tol=1e-3) and math.isclose(offset[1], 0, abs_tol=1e-3):
                stack.offset(*offset, command_line_units or MM)

            if not math.isclose(theta, 0, abs_tol=1e-2):
                stack.rotate(theta, cx, cy)

            if target is None:
                target = stack
            else:
                target.merge(stack)

    if output_board_name:
        if not output_naming_scheme:
            warnings.warn('--output-board-name given without --output-naming-scheme. This will be ignored.')
        target.board_name = output_board_name
    output_format = None if output_format == 'reuse' else FileSettings.defaults()
    target.save_to_directory(outpath, naming_scheme=output_naming_scheme or {},
                            gerber_settings=output_format,
                            excellon_settings=dataclasses.replace(output_format, zeros=None))


@cli.command()
@click.option('--version', is_flag=True, callback=_print_version, expose_value=False, is_eager=True)
@click.option('--warnings', 'format_warnings', type=click.Choice(['default', 'ignore', 'once']), default='default',
              help='''Enable or disable file format warnings during parsing (default: on)''')
@click.option('--units', type=Unit(), help='Output bounding box in this unit (default: millimeter)')
@click.option('--input-number-format', help='Override number format of input file (mostly useful for Excellon files)')
@click.option('--input-units', type=Unit(), help='Override units of input file')
@click.option('--input-zero-suppression', type=click.Choice(['off', 'leading', 'trailing']), help='Override zero suppression setting of input file')
@click.argument('infile')
def bounding_box(infile, format_warnings, input_number_format, input_units, input_zero_suppression, units):
    """ Print the bounding box of a gerber file in "[x_min] [y_min] [x_max] [y_max]" format. The bounding box contains
    all graphic objects in this file, so e.g. a 100 mm by 100 mm square drawn with a 1mm width circular aperture will
    result in an 101 mm by 101 mm bounding box.
    """

    input_settings = FileSettings()
    if input_number_format:
        a, _, b = input_number_format.partition('.')
        input_settings.number_format = (int(a), int(b))

    if input_zero_suppression:
        input_settings.zeros = None if input_zero_suppression == 'off' else input_zero_suppression

    input_settings.unit = input_units

    with warnings.catch_warnings():
        warnings.simplefilter(format_warnings)
        f = GerberFile.open(infile, override_settings=input_settings)

    (x_min, y_min), (x_max, y_max) = f.bounding_box(unit=units)
    print(f'{x_min:.6f} {y_min:.6f} {x_max:.6f} {y_max:.6f} [{units}]')


@cli.command()
@click.option('--version', is_flag=True, callback=_print_version, expose_value=False, is_eager=True)
@click.option('--warnings', 'format_warnings', type=click.Choice(['default', 'ignore', 'once']), default='default',
              help='''Enable or disable file format warnings during parsing (default: on)''')
@click.option('--force-zip', is_flag=True, help='Force treating input path as zip file (default: guess file type from extension and contents)')
@click.argument('path', type=click.Path(exists=True))
def layers(path, force_zip, format_warnings):
    """ Read layers from a directory or zip with Gerber files and list the found layer / path assignment. """ 
    with warnings.catch_warnings():
        warnings.simplefilter(format_warnings)
        if force_zip:
            stack = lyr.LayerStack.open_zip(path)
        else:
            stack = lyr.LayerStack.open(path)

    print(f'Detected board name: {stack.board_name}')
    print(f'Probably exported by: {stack.generator or "Unknown"}')
    print(f'Board bounding box: {stack.bounding_box()} [mm]')

    if stack.netlist:
        print(f'Found netlist at {stack.netlist.original_path}')
    else:
        print('No netlist found')

    print('Graphical layers:')
    for (side, function), layer in stack.graphic_layers.items():
        print(f'{side} {function}: {layer}')
    if not stack.graphic_layers:
        print('(no graphical layers)')

    print('Drill layers:')
    for layer in stack.drill_layers:
        print(layer)
    if not stack.drill_layers:
        print('(no drill layers)')


@cli.command()
@click.option('--version', is_flag=True, callback=_print_version, expose_value=False, is_eager=True)
@click.option('--warnings', 'format_warnings', type=click.Choice(['default', 'ignore', 'once']), help='''Enable or
              disable file format warnings during parsing (default: on)''')
@click.option('--force-zip', is_flag=True, help='Force treating input path as zip file (default: guess file type from extension and contents)')
@click.argument('path', type=click.Path(exists=True))
def meta(path, force_zip, format_warnings):
    """ Extract layer mapping and print it along with layer metadata as JSON to stdout. A machine-readable variant of
    the "layers" command. All lengths in the JSON are given in millimeter. """

    with warnings.catch_warnings():
        warnings.simplefilter(format_warnings)
        if force_zip:
            stack = lyr.LayerStack.open_zip(path)
        else:
            stack = lyr.LayerStack.open(path)

    out = {}
    out['board_name'] = stack.board_name
    out['generator'] = stack.generator
    (min_x, min_y), (max_x, max_y) = stack.bounding_box(default=((None, None), (None, None)))
    out['bounding_box'] = {'min_x': min_x, 'min_y': min_y, 'max_x': max_x, 'max_y': max_y}
    out['path'] = str(stack.original_path)

    if stack.netlist:
        out['netlist'] = {
                'format': 'IPC-356',
                'path': str(stack.netlist.original_path),
                'records': len(stack.netlist.test_records),
                'conductors': len(stack.netlist.conductors),
                'outlines': len(stack.netlist.outlines),
        }

    out['graphical_layers'] = {}
    for (side, function), layer in stack.graphic_layers.items():
        d = out['graphical_layers'][side] = out['graphical_layers'].get(side, {})
        (min_x, min_y), (max_x, max_y) = layer.bounding_box(default=((None, None), (None, None)))

        if layer.import_settings:
            numf = layer.import_settings.number_format
            format_settings = {
                'unit': str(layer.import_settings.unit),
                'number_format': f'{numf[0]}.{numf[1]}' if numf else None,
                'zero_suppression': str(layer.import_settings.zeros),
            }

        d[function] = {
                'format': 'Gerber',
                'path': str(layer.original_path),
                'apertures': len(list(layer.apertures())),
                'objects': len(layer.objects),
                'bounding_box': {'min_x': min_x, 'min_y': min_y, 'max_x': max_x, 'max_y': max_y},
                'format_settings': format_settings,
        }

    out['drill_layers'] = []
    for layer in stack.drill_layers:
        if layer.import_settings:
            numf = layer.import_settings.number_format
            format_settings = {
                'unit': str(layer.import_settings.unit),
                'number_format': f'{numf[0]}.{numf[1]}' if numf else None,
                'zero_suppression': str(layer.import_settings.zeros),
            }

        out['drill_layers'].append({
            'format': 'Excellon',
            'path': str(layer.original_path),
            'plating': layer.plating_type,
            'format_settings': format_settings,
        })

    print(json.dumps(out))


if __name__ == '__main__':
    cli()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2021 Jan Sebastian Götte <gerbonara@jaseg.de>

from dataclasses import dataclass, field, replace
import operator
import re
import ast
import copy
import math

from . import primitive as ap
from .expression import *
from ..utils import MM

# we make our own here instead of using math.degrees to make sure this works with expressions, too.
def rad_to_deg(x):
    return (x / math.pi) * 180

def _map_expression(node):
    if isinstance(node, ast.Num):
        return ConstantExpression(node.n)

    elif isinstance(node, ast.BinOp):
        op_map = {ast.Add: operator.add, ast.Sub: operator.sub, ast.Mult: operator.mul, ast.Div: operator.truediv}
        return OperatorExpression(op_map[type(node.op)], _map_expression(node.left), _map_expression(node.right))

    elif isinstance(node, ast.UnaryOp):
        if type(node.op) == ast.UAdd:
            return _map_expression(node.operand)
        else:
            return OperatorExpression(operator.sub, ConstantExpression(0), _map_expression(node.operand))

    elif isinstance(node, ast.Name):
        return VariableExpression(int(node.id[3:])) # node.id has format var[0-9]+

    else:
        raise SyntaxError('Invalid aperture macro expression')

def _parse_expression(expr):
    expr = expr.lower().replace('x', '*')
    expr = re.sub(r'\$([0-9]+)', r'var\1', expr)
    try:
        parsed = ast.parse(expr, mode='eval').body
    except SyntaxError as e:
        raise SyntaxError('Invalid aperture macro expression') from e
    return _map_expression(parsed)

@dataclass(frozen=True, slots=True)
class ApertureMacro:
    name: str = field(default=None, hash=False, compare=False)
    primitives: tuple = ()
    variables: tuple = ()
    comments: tuple = field(default=(), hash=False, compare=False)

    def __post_init__(self):
        if self.name is None or re.match(r'GNX[0-9A-F]{16}', self.name):
            # We can't use field(default_factory=...) here because that factory doesn't get a reference to the instance.
            self._reset_name()

    def _reset_name(self):
        object.__setattr__(self, 'name', f'GNX{hash(self)&0xffffffffffffffff:016X}')

    @classmethod
    def parse_macro(kls, macro_name, body, unit):
        comments = []
        variables = {}
        primitives = []

        blocks = body.split('*')
        for block in blocks:
            if not (block := block.strip()): # empty block
                continue

            if block.startswith('0 '): # comment
                comments.append(block[2:])
                continue
            
            block = re.sub(r'\s', '', block)

            if block[0] == '$': # variable definition
                name, expr = block.partition('=')
                number = int(name[1:])
                if number in variables:
                    raise SyntaxError(f'Re-definition of aperture macro variable {number} inside macro')
                variables[number] = _parse_expression(expr)

            else: # primitive
                primitive, *args = block.split(',')
                args = [ _parse_expression(arg) for arg in args ]
                primitives.append(ap.PRIMITIVE_CLASSES[int(primitive)].from_arglist(unit, args))

        variables = [variables.get(i+1) for i in range(max(variables.keys(), default=0))]
        return kls(macro_name, tuple(primitives), tuple(variables), tuple(comments))

    def __str__(self):
        return f'<Aperture macro {self.name}, variables {str(self.variables)}, primitives {self.primitives}>'

    def __repr__(self):
        return str(self)

    def dilated(self, offset, unit=MM):
        new_primitives = []
        for primitive in self.primitives:
            try:
                if primitive.exposure.calculate():
                    new_primitives += primitive.dilated(offset, unit)
            except IndexError:
                warnings.warn('Cannot dilate aperture macro primitive with exposure value computed from macro variable.')
                pass
        return replace(self, primitives=tuple(new_primitives))

    def to_gerber(self, unit=None):
        """ Serialize this macro's content (without the name) into Gerber using the given file unit """
        comments = [ f'0 {c.replace("*", "_").replace("%", "_")}' for c in self.comments ]
        variable_defs = [ f'${var}={str(expr)[1:-1]}' for var, expr in enumerate(self.variables, start=1) if expr is not None ]
        primitive_defs = [ prim.to_gerber(unit) for prim in self.primitives ]
        return '*\n'.join(comments + variable_defs + primitive_defs)

    def to_graphic_primitives(self, offset, rotation, parameters : [float], unit=None, polarity_dark=True):
        variables = {i: v for i, v in enumerate(self.variables, start=1) if v is not None}
        for number, value in enumerate(parameters, start=1):
            if number in variables:
                raise SyntaxError(f'Re-definition of aperture macro variable {number} through parameter {value}')
            variables[number] = value

        for primitive in self.primitives:
            yield from primitive.to_graphic_primitives(offset, rotation, variables, unit, polarity_dark)

    def rotated(self, angle):
        # aperture macro primitives use degree counter-clockwise, our API uses radians clockwise
        return replace(self, primitives=tuple(
            replace(primitive, rotation=primitive.rotation - rad_to_deg(angle)) for primitive in self.primitives))

    def scaled(self, scale):
        return replace(self, primitives=tuple(
            primitive.scaled(scale) for primitive in self.primitives))


var = VariableExpression
deg_per_rad = 180 / math.pi

class GenericMacros:

    _generic_hole = lambda n: (ap.Circle('mm', 0, var(n), 0, 0),)

    # NOTE: All generic macros have rotation values specified in **clockwise radians** like the rest of the user-facing
    # API.
    circle = ApertureMacro('GNC', (
        ap.Circle('mm', 1, var(1), 0, 0, var(4) * -deg_per_rad),
        *_generic_hole(2)))

    rect = ApertureMacro('GNR', (
        ap.CenterLine('mm', 1, var(1), var(2), 0, 0, var(5) * -deg_per_rad),
        *_generic_hole(3)))

    # params: width, height, corner radius, *hole, rotation
    rounded_rect = ApertureMacro('GRR', (
        ap.CenterLine('mm', 1, var(1)-2*var(3), var(2), 0, 0, var(6) * -deg_per_rad),
        ap.CenterLine('mm', 1, var(1), var(2)-2*var(3), 0, 0, var(6) * -deg_per_rad),
        ap.Circle('mm', 1, var(3)*2, +(var(1)/2-var(3)), +(var(2)/2-var(3)), var(6) * -deg_per_rad),
        ap.Circle('mm', 1, var(3)*2, +(var(1)/2-var(3)), -(var(2)/2-var(3)), var(6) * -deg_per_rad),
        ap.Circle('mm', 1, var(3)*2, -(var(1)/2-var(3)), +(var(2)/2-var(3)), var(6) * -deg_per_rad),
        ap.Circle('mm', 1, var(3)*2, -(var(1)/2-var(3)), -(var(2)/2-var(3)), var(6) * -deg_per_rad),
        *_generic_hole(4)))

    # params: width, height, length difference between narrow side (top) and wide side (bottom), *hole, rotation
    isosceles_trapezoid = ApertureMacro('GTR', (
        ap.Outline('mm', 1, 4,
                          (var(1)/-2,            var(2)/-2,
                          var(1)/-2+var(3)/2,   var(2)/2,
                          var(1)/2-var(3)/2,    var(2)/2,
                          var(1)/2,             var(2)/-2,
                          var(1)/-2,            var(2)/-2,),
                          var(6) * -deg_per_rad),
        *_generic_hole(4)))

    # params: width, height, length difference between narrow side (top) and wide side (bottom), margin, *hole, rotation
    rounded_isosceles_trapezoid = ApertureMacro('GRTR', (
        ap.Outline('mm', 1, 4,
                          (var(1)/-2,            var(2)/-2,
                          var(1)/-2+var(3)/2,   var(2)/2,
                          var(1)/2-var(3)/2,    var(2)/2,
                          var(1)/2,             var(2)/-2,
                          var(1)/-2,            var(2)/-2,),
                          var(6) * -deg_per_rad),
        ap.VectorLine('mm', 1, var(4)*2, 
                          var(1)/-2,            var(2)/-2,
                          var(1)/-2+var(3)/2,   var(2)/2,),
        ap.VectorLine('mm', 1, var(4)*2, 
                          var(1)/-2+var(3)/2,   var(2)/2,
                          var(1)/2-var(3)/2,    var(2)/2,),
        ap.VectorLine('mm', 1, var(4)*2, 
                          var(1)/2-var(3)/2,    var(2)/2,
                          var(1)/2,             var(2)/-2,),
        ap.VectorLine('mm', 1, var(4)*2, 
                          var(1)/2,             var(2)/-2,
                          var(1)/-2,            var(2)/-2,),
        ap.Circle('mm', 1, var(4)*2, 
                          var(1)/-2,            var(2)/-2,),
        ap.Circle('mm', 1, var(4)*2, 
                          var(1)/-2+var(3)/2,   var(2)/2,),
        ap.Circle('mm', 1, var(4)*2, 
                          var(1)/2-var(3)/2,    var(2)/2,),
        ap.Circle('mm', 1, var(4)*2, 
                          var(1)/2,             var(2)/-2,),
        *_generic_hole(5)))

    # w must be larger than h
    # params: width, height, *hole, rotation
    obround = ApertureMacro('GNO', (
        ap.CenterLine('mm', 1, var(1)-var(2), var(2), 0, 0, var(5) * -deg_per_rad),
        ap.Circle('mm', 1, var(2), +(var(1)-var(2))/2, 0, var(5) * -deg_per_rad),
        ap.Circle('mm', 1, var(2), -(var(1)-var(2))/2, 0, var(5) * -deg_per_rad),
        *_generic_hole(3) ))

    polygon = ApertureMacro('GNP', (
        ap.Polygon('mm', 1, var(2), 0, 0, var(1), var(3) * -deg_per_rad),
        ap.Circle('mm', 0, var(4), 0, 0)))


if __name__ == '__main__':
    import sys
    #for line in sys.stdin:
        #expr = _parse_expression(line.strip())
        #print(expr, '->', expr.optimized())

    for primitive in parse_macro(sys.stdin.read(), 'mm'):
        print(primitive)




#!/usr/bin/env python

from pathlib import Path
import unicodedata
import re
import ast
from functools import lru_cache
import math
from importlib.resources import files

from . import data
from .utils import rotate_point, Tag


STROKE_FONT_SCALE = 1/21
FONT_OFFSET = -10
DEFAULT_SPACE_WIDTH = 0.6
DEFAULT_CHAR_GAP = 0.2

_dec = lambda c: ord(c)-ord('R') 


class Newstroke:
    def __init__(self, newstroke_cpp=None):
        if newstroke_cpp is None:
            newstroke_cpp = files(data).joinpath('newstroke_font.cpp').read_bytes()
        self.glyphs = dict(self.load_font(newstroke_cpp))

    @classmethod
    @lru_cache
    def load(kls):
        return kls()

    def render(self, text, size=1.0, x0=0, y0=0, rotation=0, h_align='left', v_align='bottom', space_width=DEFAULT_SPACE_WIDTH, char_gap=DEFAULT_CHAR_GAP, scale=(1, 1), mirror=(False, False)):
        text = unicodedata.normalize('NFC', text)
        missing_glyph = self.glyphs['?']
        sx, sy = scale
        mx, my = mirror
        x = 0

        if rotation >= 180:
            rotation -= 180
            h_align = {'left': 'right', 'right': 'left'}.get(h_align, h_align)
            x0, y0 = -x0, y0

#        if mx:
#            y0 = -y0
#            if rotation == 0:
#                v_align = {'top': 'bottom', 'bottom': 'top'}.get(v_align, v_align)
#            else:
#                h_align = {'left': 'right', 'right': 'left'}.get(h_align, h_align)

        x0, y0 = rotate_point(x0, y0, math.radians(-rotation))

        alx, aly = 0, 0
        (minx, miny), (maxx, maxy) = bbox = self.bounding_box(text, size, space_width, char_gap)
        w = maxx - minx

        if my:
            if rotation == 0:
                sx = -1
                h_align = {'left': 'right', 'right': 'left'}.get(h_align, h_align)
            else:
                sy = -sy

        if h_align != 'left':
            if h_align == 'right':
                alx = -w
            elif h_align == 'center':
                alx = -w/2
            else:
                raise ValueError(f'Invalid h_align value "{h_align}"')

        if v_align == 'top':
            aly = sy*1.2*size
        elif v_align == 'middle':
            aly = sy*1.2*size/2
        elif v_align != 'bottom':
                raise ValueError(f'Invalid v_align value "{v_align}"')

        for c in text:
            if c == ' ':
                x += space_width
                continue

            width, strokes = self.glyphs.get(c, missing_glyph)
            glyph_w = max(width, max(x for st in strokes for x, _y in st))

            for st in strokes:
                yield [rotate_point((px+x)*sx*size+alx+x0, py*sy*size+aly+y0, math.radians(-rotation), x0, y0) for px, py in st]

            x += glyph_w

    def render_svg(self, text, size=1.0, x0=0, y0=0, rotation=0, h_align='left', v_align='bottom', space_width=DEFAULT_SPACE_WIDTH, char_gap=DEFAULT_CHAR_GAP, scale=(1, -1), mirror=(False, False), **svg_attrs):
        if 'stroke_linecap' not in svg_attrs:
            svg_attrs['stroke_linecap'] = 'round'
        if 'stroke_linejoin' not in svg_attrs:
            svg_attrs['stroke_linejoin'] = 'round'
        if 'stroke_width' not in svg_attrs:
            svg_attrs['stroke_width'] = f'{0.2*size:.3f}'
        svg_attrs['fill'] = 'none'

        strokes = ['M ' + ' L '.join(f'{x:.3f} {y:.3f}' for x, y in stroke)
                   for stroke in self.render(text, size=size, x0=x0, y0=y0, rotation=rotation, h_align=h_align,
                                             v_align=v_align, mirror=mirror, space_width=space_width, char_gap=char_gap,
                                             scale=scale)]
        return Tag('path', d=' '.join(strokes), **svg_attrs)

    def bounding_box(self, text, size=1.0, space_width=DEFAULT_SPACE_WIDTH, char_gap=DEFAULT_CHAR_GAP):
        text = unicodedata.normalize('NFC', text)
        missing_glyph = self.glyphs['?']
        x = 0
        for c in text:
            if c == ' ':
                x += space_width*size
                continue

            width, strokes = self.glyphs.get(c, missing_glyph)
            glyph_w = max(width, max(x for st in strokes for x, _y in st))
            x += glyph_w*size

        return (0, -0.2*size), (x, 1.2*size)

    def load_font(self, newstroke_cpp):
        e = []
        for char, (width, strokes) in self.load_glyphs(newstroke_cpp):
            yield char, (width, strokes)

    @classmethod
    def decode_stroke(kls, stroke, start_x):
        for i in range(0, len(stroke), 2):
            x = (stroke[i]-0x52-start_x)*STROKE_FONT_SCALE
            y = (stroke[i+1]-0x52+FONT_OFFSET)*STROKE_FONT_SCALE
            yield (x, y)

    @classmethod
    def decode_glyph(kls, data):
        start_x, end_x = data[0]-0x52, data[1]-0x52
        width = end_x - start_x

        strokes = tuple(tuple(kls.decode_stroke(st, start_x)) for st in data[2:].split(b' R'))
        return width*STROKE_FONT_SCALE, strokes

    @classmethod
    def load_glyphs(kls, newstroke_cpp):
        it = iter(newstroke_cpp.splitlines())
        
        for line in it:
            if re.search(rb'char.*\*', line):
                break

        charcode = 0x20
        for line in it:
            if (match := re.search(rb'".*"', line)):
                yield chr(charcode), kls.decode_glyph(match.group(0)[1:-1].replace(b'\\\\', b'\\'))
                charcode += 1
            else:
                if b'}' in line:
                    break


if __name__ == '__main__':
    import time
    t1 = time.time()
    Newstroke()
    t2 = time.time()
    print((t2-t1)*1000)
